
/*!
 * \example uart/slmain.c
 *
 * This sample demonstrates the usage of the ATmega 128 on-chip UART.
 * Note, that we don't do any error checking, because without this
 * UART we can't tell the user our problem.
 */

#include <string.h>

#include <dev/uartavr.h>    /* Reference to the uart0 device driver. */

#include <sys/thread.h>     /* Contains the THREAD macro. */

#include <sys/print.h>      /* Formatted output functions. */

static char *banner = "\r\nSL UART Sample\r\n";

static char inbuf[128];

/*! \fn slMain(void *arg)
 * \brief Main application routine. 
 *
 * sl/OS automatically calls this entry after initialization.
 */
#include <sys/timer.h>
THREAD(slMain, arg)
{
    int got;
    int i;
    NUTDEVICE *uart;
    u_long baud = 115200;

    /*
     * Each device must be registered. We do this
     * by referencing the device structure of the
     * driver. The advantage is, that only those
     * device drivers are included in our flash
     * code, which we really need.
     *
     * The uart0 device is the one on the ATmega
     * chip. So it has no configurable base address
     * or interrupt and we set both items to zero.
     */
    slRegisterDevice(&devUart0, 0, 0);

    /*
     * Now, as the device is registered, we can
     * open it. On stream devices, like uart0,
     * this call will initialize the I/O buffers.
     * The function returns a pointer to the
     * device structure, which we use for subsequent
     * reading and writing.
     */
    uart = slDeviceOpen("uart0");
    slDeviceIOCtl(uart, UART_SETSPEED, &baud);

    /*
     * Stream devices can use simple read and write
     * functions. Writing program space data is
     * supported too.
     */
    slDeviceWrite(uart, banner, strlen(banner));
	{
	    static const char pk[] = "Press key...";
        slDeviceWrite_P(uart, pk, 12);
	}

    /*
     * Stream devices do buffered I/O. That means,
     * nothing will be passed to the hardware
     * device until either the output buffer is
     * full or we do a flush. Writing nothing
     * will flush the output buffer.
     */
    slDeviceWrite(uart, 0, 0);

    /*
     * The slDeviceRead() function will grab all
     * available bytes from the input buffer.
     * If the buffer is empty, the call will
     * block until something is available
     * for reading.
     */
    got = slDeviceRead(uart, inbuf, sizeof(inbuf));
    slDeviceWrite(uart, inbuf, got);

    /*
     * sl/OS never expects a thread to return.
     * So we enter an endless loop here.
     */
    for(i = 0;; i++) {
        /*
         * A bit more advanced input routine is
         * able to read a string up to and including
         * the first newline character or until a
         * specified maximum number of characters,
         * whichever comes first.
         */
        slDeviceWrite(uart, "\r\nEnter your name: ", 19);
        slDeviceWrite(uart, 0, 0);
        got = slDeviceGetLine(uart, inbuf, sizeof(inbuf) - 1);
        inbuf[got] = 0;

        /*
         * Stream devices can use the more advanced
         * slPrint functions. They support formatted
         * output as well as printing strings from
         * program space.
         */
        if(inbuf[0])
            slPrintFormat(uart, "Hello %s!\r\n", inbuf);
        else {
		    static const char ss[] = "Hello stranger!\r\n";
            slPrintString_P(uart, ss);
        }
		
        /*
         * We might have used slDeviceWrite() to flush
         * the output buffer. However, slPrintFlush
         * looks nicer.
         */
        slPrintFlush(uart);

        /*
         * Simply to demonstrate some formatting
         * options.
         */
        slPrintFormat(uart, "Loop +%udec %04Xhex\r\n", i + 1, i + 1);
    }
}
