
#include <string.h>

#include <sys/heap.h>
#include <dev/netbuf.h>


static int slNetBufAllocData(NBDATA *nbd, u_short size)
{
    if((nbd->vp = slHeapAlloc(size)) == 0) {
        nbd->sz = 0;
        return -1;
    }
    nbd->sz = size;
    return 0;
}

static void slNetBufFreeData(NBDATA *nbd)
{
    slHeapFree(nbd->vp);
    nbd->vp = 0;
    nbd->sz = 0;
}

/*!
 * \brief Allocate or re-allocate a network buffer part.
 *
 * \param nb   Points to an existing network buffer structure
 *             or NULL, if a new structure should be created.
 * \param type Part of the buffer to be allocated.
 * \param size Size of the part to be allocated.
 *
 * \return Pointer to the allocated network buffer
 *         structure.
 */
NETBUF *slNetBufAlloc(NETBUF *nb, u_char type, u_short size)
{
    if(nb == 0) {
        nb = slHeapAllocClear(sizeof(NETBUF));
    }

    if(nb && type) {
        if(size) {
            if(type & NBAF_DATALINK) {
                if(nb->nb_flags & NBAF_DATALINK) {
                    if(nb->nb_dl.sz < size) {
                        slNetBufFreeData(&nb->nb_dl);
                        if(slNetBufAllocData(&nb->nb_dl, size)) {
                            slNetBufFree(nb);
                            return 0;
                        }
                    }
                }
                else if(slNetBufAllocData(&nb->nb_dl, size)) {
                    slNetBufFree(nb);
                    return 0;
                }
            }

            if(type & NBAF_NETWORK) {
                if(nb->nb_flags & NBAF_NETWORK) {
                    if(nb->nb_nw.sz < size) {
                        slNetBufFreeData(&nb->nb_nw);
                        if(slNetBufAllocData(&nb->nb_nw, size)) {
                            slNetBufFree(nb);
                            return 0;
                        }
                    }
                }
                else if(slNetBufAllocData(&nb->nb_nw, size)) {
                    slNetBufFree(nb);
                    return 0;
                }
            }

            if(type & NBAF_TRANSPORT) {
                if(nb->nb_flags & NBAF_TRANSPORT) {
                    if(nb->nb_tp.sz < size) {
                        slNetBufFreeData(&nb->nb_tp);
                        if(slNetBufAllocData(&nb->nb_tp, size)) {
                            slNetBufFree(nb);
                            return 0;
                        }
                    }
                }
                else if(slNetBufAllocData(&nb->nb_tp, size)) {
                    slNetBufFree(nb);
                    return 0;
                }
            }

            if(type & NBAF_APPLICATION) {
                if(nb->nb_flags & NBAF_APPLICATION) {
                    if(nb->nb_ap.sz < size) {
                        slNetBufFreeData(&nb->nb_ap);
                        if(slNetBufAllocData(&nb->nb_ap, size)) {
                            slNetBufFree(nb);
                            return 0;
                        }
                    }
                }
                else if(slNetBufAllocData(&nb->nb_ap, size)) {
                    slNetBufFree(nb);
                    return 0;
                }
            }
            nb->nb_flags |= type;
        }
        else {
            type &= nb->nb_flags;
            if(type & NBAF_DATALINK)
                slNetBufFreeData(&nb->nb_dl);
            if(type & NBAF_NETWORK)
                slNetBufFreeData(&nb->nb_nw);
            if(type & NBAF_TRANSPORT)
                slNetBufFreeData(&nb->nb_tp);
            if(type & NBAF_APPLICATION)
                slNetBufFreeData(&nb->nb_ap);
            nb->nb_flags &= ~type;
        }
    }
    return nb;
}

/*!
 * \brief Create a copy of an existing network buffer
 *        structure.
 *
 * \param nb Points to an existing network buffer
 *           structure, previously allocated by
 *           slNetBufAlloc().
 *
 * \return Pointer to a newly allocated copy.
 */
NETBUF *slNetBufClone(NETBUF *nb)
{
    NETBUF *clone;

    if((clone = slHeapAllocClear(sizeof(NETBUF))) == 0)
        return 0;

    if(nb->nb_dl.sz) {
        if(slNetBufAllocData(&clone->nb_dl, nb->nb_dl.sz)) {
            slNetBufFree(clone);
            return 0;
        }
        memcpy(clone->nb_dl.vp, nb->nb_dl.vp, nb->nb_dl.sz);
        clone->nb_flags |= NBAF_DATALINK;
    }
    if(nb->nb_nw.sz) {
        if(slNetBufAllocData(&clone->nb_nw, nb->nb_nw.sz)) {
            slNetBufFree(clone);
            return 0;
        }
        memcpy(clone->nb_nw.vp, nb->nb_nw.vp, nb->nb_nw.sz);
        clone->nb_flags |= NBAF_NETWORK;
    }
    if(nb->nb_tp.sz) {
        if(slNetBufAllocData(&clone->nb_tp, nb->nb_tp.sz)) {
            slNetBufFree(clone);
            return 0;
        }
        memcpy(clone->nb_tp.vp, nb->nb_tp.vp, nb->nb_tp.sz);
        clone->nb_flags |= NBAF_TRANSPORT;
    }
    if(nb->nb_ap.sz) {
        if(slNetBufAllocData(&clone->nb_ap, nb->nb_ap.sz)) {
            slNetBufFree(clone);
            return 0;
        }
        memcpy(clone->nb_ap.vp, nb->nb_ap.vp, nb->nb_ap.sz);
        clone->nb_flags |= NBAF_APPLICATION;
    }
    return clone;
}

/*!
 * \brief Release a network buffer structure.
 *
 * Returns all memory previously allocated by a
 * network buffer to the available heap space.
 *
 * \param nb Points to an existing network buffer
 *           structure, previously allocated by
 *           slNetBufAlloc().
 *
 * \return 0 if successfull or -1 if the structure
 *         contains previously released memory space.
 */
int slNetBufFree(NETBUF *nb)
{
    slNetBufAlloc(nb, nb->nb_flags, 0);
    if(slHeapFree(nb)) {
        return -1;
    }
    return 0;
}

/*@}*/
