
#include <string.h>

#include <sys/heap.h>
#include <fs/uromfs.h>

/*!
 * \addtogroup xgurom
 */
/*@{*/


/*!
 * \brief Opens an existing file for reading.
 *
 * \param name Points to a string that specifies the name of the
 *             file to open. The name must exactly match the
 *             full pathname of the file.
 *
 * \return A pointer to a ROMFILE structure that can be used
 *         to read from the file and retrieve information about
 *         the file.
 */
ROMFILE *slRomFileOpen(char *name)
{
    ROMENTRY *rome;
    ROMFILE *romf = 0;

    for(rome = romEntryList; rome; rome = rome->rome_next) {
        if(strcmp(rome->rome_name, name) == 0)
            break;
    }
    if(rome) {
        if((romf = slHeapAllocClear(sizeof(ROMFILE))) != 0)
            romf->romf_entry = rome;
    }
    return romf;
}

/*!
 * \brief Close a previously opened file.
 *
 * \param romf Identifies the file to close. This pointer must
 *             have been created by calling slRomFileOpen().
 *
 * \return 0 if the function is successful, -1 otherwise.
 */
int slRomFileClose(ROMFILE *romf)
{
    return slHeapFree(romf);
}

/*!
 * \brief Read data from a file.
 *
 * Read up to a specified number of bytes of data from a
 * file into a buffer. The function may read fewer than
 * the specified number of bytes if it reaches the end of
 * the file.
 *
 * \param romf Identifies the file to read from. This pointer must
 *             have been created by calling slRomFileOpen().
 * \param data Points to the buffer that receives the data.
 * \param size Specifies the number of bytes to read from the file.
 *
 * \return The number of bytes read from the file or -1 if
 *         an error occured.
 */
int slRomFileRead(ROMFILE *romf, void *data, int size)
{
    ROMENTRY *rome = romf->romf_entry;

    if(size > rome->rome_size - romf->romf_pos)
        size = rome->rome_size - romf->romf_pos;
    if(size) {
        memcpy_P(data, rome->rome_data + romf->romf_pos, size);
        romf->romf_pos += size;
    }
    return size;
}

/*!
 * \brief Retrieve the size of a file.
 *
 * \param romf Identifies the file to query. This pointer must
 *             have been created by calling slRomFileOpen().
 *
 * \return The number of bytes in this file or -1 if an
 *         error occured.
 */
int slRomFileSize(ROMFILE *romf)
{
    return romf->romf_entry->rome_size;
}

/*!
 * \brief Move the file pointer to a new position.
 *
 * The file pointer is maintained by sl/OS. It points to
 * the next byte to be read from a file. The file pointer
 * is automatically incremented for each byte read. When
 * the file is opened, it is at position 0, the beginning
 * of the file.
 *
 * \param romf Identifies the file to seek. This pointer must
 *             have been created by calling slRomFileOpen().
 * \param pos  Specifies the new absolute position of the file pointer.
 *
 * \return 0 if the function is successful, -1 otherwise.
 */
int slRomFileSeek(ROMFILE *romf, int pos)
{
    if(pos < 0 || pos > slRomFileSize(romf))
        return -1;
    romf->romf_pos = pos;

    return 0;
}

/*@}*/
