#ifndef _NETINET_IP_ICMP_H_
#define _NETINET_IP_ICMP_H_


#include <sys/types.h>
#include <net/if_arp.h>
#include <netinet/ip.h>

/*!
 * \file netinet/ip_icmp.h
 * \brief Interface control message protocol definitions.
 *
 * RFC 792.
 */

/*
 * Definition of type and code field values.
 */
#define ICMP_ECHOREPLY                  0   /*!< \brief Echo reply type. */
#define ICMP_UNREACH                    3   /*!< \brief Destination unreachable type. */
#define     ICMP_UNREACH_NET            0   /*!< \brief Destination unreachable: Bad net code. */
#define     ICMP_UNREACH_HOST           1   /*!< \brief Destination unreachable: Bad host. */
#define     ICMP_UNREACH_PROTOCOL       2   /*!< \brief Destination unreachable: Bad protocol. */
#define     ICMP_UNREACH_PORT           3   /*!< \brief Destination unreachable: Bad port. */
#define     ICMP_UNREACH_NEEDFRAG       4   /*!< \brief Destination unreachable: Fragmentation needed. */
#define     ICMP_UNREACH_SRCFAIL        5   /*!< \brief Destination unreachable: Source route failed. */
#define     ICMP_UNREACH_NET_UNKNOWN    6   /*!< \brief Destination unreachable: Unknown net. */
#define     ICMP_UNREACH_HOST_UNKNOWN   7   /*!< \brief Destination unreachable: Unknown host. */
#define     ICMP_UNREACH_ISOLATED       8   /*!< \brief Destination unreachable: Source host isolated. */
#define     ICMP_UNREACH_NET_PROHIB     9   /*!< \brief Destination unreachable: Net access prohibited. */
#define     ICMP_UNREACH_HOST_PROHIB    10  /*!< \brief Destination unreachable: Host access prohibited. */
#define     ICMP_UNREACH_TOSNET         11  /*!< \brief Destination unreachable: Bad tos for net. */
#define     ICMP_UNREACH_TOSHOST        12  /*!< \brief Destination unreachable: Bad tos for host. */
#define ICMP_SOURCEQUENCH               4   /*!< \brief Packet lost type. */
#define ICMP_REDIRECT                   5   /*!< \brief Shorter route type. */
#define     ICMP_REDIRECT_NET           0   /*!< \brief Shorter route for network. */
#define     ICMP_REDIRECT_HOST          1   /*!< \brief Shorter route for host. */
#define     ICMP_REDIRECT_TOSNET        2   /*!< \brief Shorter route for tos and net. */
#define     ICMP_REDIRECT_TOSHOST       3   /*!< \brief Shorter route for tos and host. */
#define ICMP_ECHO                       8   /*!< \brief Echo reply type. */
#define ICMP_ROUTERADVERT               9   /*!< \brief Router advertisement type. */
#define ICMP_ROUTERSOLICIT              10  /*!< \brief Router solicitation type. */
#define ICMP_TIMXCEED                   11  /*!< \brief Time exceeded type. */
#define     ICMP_TIMXCEED_INTRANS       0   /*!< \brief Time exceeded in transit. */
#define     ICMP_TIMXCEED_REASS         1   /*!< \brief Time exceeded in reassembly. */
#define ICMP_PARAMPROB                  12  /*!< \brief Bad IP header type. */
#define     ICMP_PARAMPROB_OPTABSENT    1   /*!< \brief Bad IP header: Required option missing. */
#define ICMP_TSTAMP                     13  /*!< \brief Timestamp request type. */
#define ICMP_TSTAMPREPLY                14  /*!< \brief Timestamp reply type. */
#define ICMP_IREQ                       15  /*!< \brief Information request type. */
#define ICMP_IREQREPLY                  16  /*!< \brief Information reply type. */
#define ICMP_MASKREQ                    17  /*!< \brief Address mask request type. */
#define ICMP_MASKREPLY                  18  /*!< \brief Address mask reply type. */

#define ICMP_MAXTYPE                    18  /*!< \brief Maximum ICMP type number. */


#define ICMP_MINLEN 8   /*!< \brief Lower bounds on packet length. */

/*!
 * \struct icmp ip_icmp.h netinet/ip_icmp.h
 * \brief ICMP protocol header structure.
 */
/*!
 * \typedef ICMPHDR
 * \brief ICMP protocol header type.
 */
typedef struct icmp {
    u_char  icmp_type;      /*!< \brief Type of message. */
    u_char  icmp_code;      /*!< \brief Type sub code. */
    u_short icmp_cksum;     /*!< \brief Ones complement header checksum. */
    u_long  icmp_spec;      /*!< \brief Type specific information. */
} ICMPHDR;

#endif
