#ifndef _NETINET_TCP_H_
#define _NETINET_TCP_H_

#include <sys/types.h>

/*!
 * \file netinet/tcp.h
 * \brief TCP protocol definitions.
 */

#define TH_FIN  0x01    /*!< \brief Finishing transmission. */
#define TH_SYN  0x02    /*!< \brief Synchronizing sequence numbers. */
#define TH_RST  0x04    /*!< \brief Reset connection. */
#define TH_PUSH 0x08    /*!< \brief Push data to application level. */
#define TH_ACK  0x10    /*!< \brief Acknowledge field is valid. */
#define TH_URG  0x20    /*!< \brief Urgent data present. */

/*! \brief TCP flag mask. */
#define TH_FLAGS (TH_FIN | TH_SYN | TH_RST | TH_ACK | TH_URG)

/*!
 * \struct tcphdr tcp.h netinet/tcp.h
 * \brief TCP protocol header structure.
 */
/*!
 * \typedef TCPHDR
 * \brief TCP protocol header type.
 */
typedef struct tcphdr {
    u_short th_sport;       /*!< \brief Source port. */
    u_short th_dport;       /*!< \brief Destination port. */
    u_long  th_seq;         /*!< \brief Sequence number of first octet in this segment. */
    u_long  th_ack;         /*!< \brief Expected sequence number of next octet. */
#ifdef __IMAGECRAFT__
    unsigned th_x2:4,
             th_off:4;
#else
    u_char  th_x2:4,        /*!< \brief Unused. */
            th_off:4;       /*!< \brief Data offset. */
#endif
    u_char  th_flags;       /*!< \brief Control flags. */
    u_short th_win;         /*!< \brief Number of acceptable octects. */
    u_short th_sum;         /*!< \brief 96 byte pseudo header checksum. */
    u_short th_urp;         /*!< \brief Urgent data pointer. */
} TCPHDR;

#define TCPOPT_EOL                  0   /*!< \brief End of options. */
#define TCPOPT_NOP                  1   /*!< \brief Nothing. */
#define TCPOPT_MAXSEG               2   /*!< \brief Maximum segment size. */
#define    TCPOLEN_MAXSEG           4   /*!< \brief Maximum segment size length. */
#define TCPOPT_WINDOW               3   /*!< \brief Receive window. */
#define    TCPOLEN_WINDOW           3   /*!< \brief Receive window length. */

/*!
 * \struct _TCPPSEUDOHDR tcp.h netinet/tcp.h
 * \brief TCP pseudo header structure.
 */
/*!
 * \typedef TCPPSEUDOHDR
 * \brief TCP pseudo header type.
 */
typedef struct _TCPPSEUDOHDR {
    u_long  tph_src;    /*!< \brief IP address of sender. */
    u_long  tph_dst;    /*!< \brief IP address of target. */
    u_char  tph_mbz;    /*!< \brief mbz */
    u_char  tph_p;      /*!< \brief p */
    u_short tph_len;    /*!< \brief len */
    u_short tph_sum;    /*!< \brief Checksum */
} TCPPSEUDOHDR;


#define TCP_MSS             536     /*!< \brief Default maximum segment size.
                                         The maximum size of an IP datagram, that will
                                         not become fragmented, is 576. The maximum
                                         IP datagram for Ethernet is 1500. Reduce this
                                         number by 40, 20 bytes TCP header and 20 bytes
                                         IP header.
                                     */
#define TCP_WINSIZE         3216    /*!< \brief Default window size.
                                         It's recommended to set this 6 times the maximum
                                         segment size.
                                     */

#define TCP_MAXWIN          65535   /*!< \brief Largest value for (unscaled) window. */
#define TTCP_CLIENT_SND_WND 4096    /*!< \brief Default send window for T/TCP client. */

#define TCP_MAX_WINSHIFT    14      /*!< \brief Maximum window shift. */

#define TCP_MAXHLEN         (0xf<<2)    /*!< \brief Maximum length of header in bytes. */
#define TCP_MAXOLEN         (TCP_MAXHLEN - sizeof(struct tcphdr))   /*!< \brief Maximum space left for options. */

/*
 * User-settable options.
 */
#define TCP_NODELAY 0x01    /*!< \brief Don't delay send to coalesce segments. */
#define TCP_MAXSEG  0x02    /*!< \brief Set maximum segment size. */
#define TCP_NOPUSH  0x04    /*!< \brief Don't push last block of write. */
#define TCP_NOOPT   0x08    /*!< \brief Don't use TCP options. */

#endif
