#ifndef _SYS_DEVICE_H_
#define _SYS_DEVICE_H_


#include <pgmspace.h>

#include <sys/types.h>

/*!
 * \file sys/device.h
 * \brief sl/OS device definitions.
 */

/*!
 * \addtogroup xgDevice
 */
/*@{*/


// wait times for emulation and reality
// has to be overworked

#ifndef __EMULATION__
#define WAIT5		5
#define WAIT50		50
#define WAIT100		100
#define WAIT250		250
#define WAIT500		500
#else
#define WAIT5		1
#define WAIT50		5
#define WAIT100		10
#define WAIT250		25
#define WAIT500		50
#endif




#define IFTYP_RAM       0   /*!< \brief RAM device */
#define IFTYP_ROM       1   /*!< \brief ROM device */
#define IFTYP_STREAM    2   /*!< \brief Stream device */
#define IFTYP_NET       3   /*!< \brief Net device */

/*!
 * \brief Device structure type.
 */
typedef struct _NUTDEVICE NUTDEVICE;

/*!
 * \struct _NUTDEVICE device.h sys/device.h
 * \brief Device structure.
 */
struct _NUTDEVICE {
    NUTDEVICE *dev_next;       /*!< \brief Link to the next device structure. */
    u_char  dev_name[9];    /*!< \brief Unique device name. */
    u_char  dev_type;       /*!< \brief Type of interface. 
                             * May be any of the following:
                             * - IFTYP_RAM
                             * - IFTYP_ROM
                             * - IFTYP_STREAM
                             * - IFTYP_NET
                             */
    u_short dev_base;       /*!< \brief Hardware base address. 
                             *   Will be set by calling slRegisterDevice(). 
                             *   On some device drivers this address may
                             *   be fixed.
                             */
    u_char  dev_irq;        /*!< \brief Interrupt registration number. 
                             *   Will be set by calling slRegisterDevice(). 
                             *   On some device drivers the interrupt may
                             *   be fixed.
                             */
    void   *dev_icb;        /*!< \brief Interface control block.
                             *   With stream devices, this points to the
                             *   IFSTREAM structure and with network
                             *   devices this is a pointer to the IFNET
                             *   structure.
                             */
    void   *dev_dcb;        /*!< \brief Driver control block.
                             *   Points to a device specific information block.
                             */
    int   (*dev_init)(NUTDEVICE *);   /*!< \brief Driver initialization routine. 
                                    *   With stream devices this is called during
                                    *   slDeviceOpen(). For network devices
                                    *   this routine is called within slNetIfConfig().
                                    */
    int   (*dev_ioctl)(NUTDEVICE *, int, void *);  /*!< \brief Driver control function. 
                                                 *   Used to modify or query device 
                                                 *   specific settings.
                                                 */
};
/*@}*/


/*!
 * \addtogroup xgUARTIOCTL
 */
/*@{*/

#define UART_SETSPEED       0x0101  /*!< \brief Set baudrate. */
#define UART_GETSPEED       0x0102  /*!< \brief Query baudrate. */
#define UART_SETDATABITS    0x0103  /*!< \brief Set number of data bits. */
#define UART_GETDATABITS    0x0104  /*!< \brief Set number of data bits. */
#define UART_SETPARITY      0x0105  /*!< \brief Set parity. */
#define UART_GETPARITY      0x0106  /*!< \brief Query parity. */
#define UART_SETSTOPBITS    0x0107  /*!< \brief Set number of stop bits. */
#define UART_GETSTOPBITS    0x0108  /*!< \brief Query number of stop bits. */
#define UART_SETSTATUS      0x0109  /*!< \brief Set status. */
#define UART_GETSTATUS      0x010a  /*!< \brief Query status. */
#define UART_SETREADTIMEOUT 0x010b  /*!< \brief Set read timeout. */
#define UART_GETREADTIMEOUT 0x010c  /*!< \brief Query read timeout. */
#define UART_SETWRITETIMEOUT 0x010d /*!< \brief Set write timeout. */
#define UART_GETWRITETIMEOUT 0x010e /*!< \brief Query write timeout. */

#define DIO_SETSTATUS       0x0209  /*!< \brief Set digital output status. */
#define DIO_GETSTATUS       0x020a  /*!< \brief Query digital input status. */
#define DIO_WAITSTATUS      0x020b  /*!< \brief Wait for digital input status change. */

/*@}*/

/*!
 * \addtogroup xgUARTStatus
 */
/*@{*/
#define UART_FRAMINGERROR   0x00000001  /*!< \brief Framing error. */
#define UART_OVERRUNERROR   0x00000002  /*!< \brief Overrun error. */
#define UART_RXBUFFEREMPTY  0x00000004  /*!< \brief Receiver buffer empty. */
#define UART_TXBUFFEREMPTY  0x00000008  /*!< \brief Transmitter buffer empty. */
/*@}*/

extern NUTDEVICE *nutDeviceList;

extern int slRegisterDevice(NUTDEVICE *dev, u_short base, u_char irq);
extern NUTDEVICE *slDeviceLookup(CONST char *name);
extern NUTDEVICE *slDeviceOpen(CONST char *name);
extern int slDeviceClose(NUTDEVICE *dev);
extern int slDeviceRead(NUTDEVICE *dev, void *data, int size);
extern int slDeviceReadTran(NUTDEVICE *dev, void *data, int size);
extern int slDeviceGetLine(NUTDEVICE *dev, void *data, int size);
extern int slDeviceWrite(NUTDEVICE *dev, CONST void *data, int len);
extern int slDeviceWriteTran(NUTDEVICE *dev, CONST void *data);
extern int slDeviceWrite_P(NUTDEVICE *dev, PGM_P data, int len);
extern int slDeviceIOCtl(NUTDEVICE *dev, int req, void *conf);

#ifdef __GNUC__
#ifdef __AVR_ATmega128__
#define UDR     UDR0
#define UBRR    UBRR0L
#define USR     UCSR0A
#define UCR     UCSR0B
#define EICR    EICRB
#endif

#else

#ifdef __AVR_ATmega128__
#ifndef UDR
#define UDR     UDR0
#define UBRR    UBRR0L
#define USR     UCSR0A
#define UCR     UCSR0B
#endif

#ifndef RXC
#define RXC     RXC0
#define UDRE    UDRE0
#define FE      FE0
#define OVR     3
#define RXCIE   RXCIE0
#define TXCIE   TXCIE0
#define RXEN    RXEN0
#define TXEN    TXEN0
#define EICR    EICRB
#endif
#endif

#endif

#endif
