
#include <string.h>

#include <sys/atom.h>
#include <sys/heap.h>
#include <sys/event.h>
#include <sys/thread.h>
#include <sys/timer.h>

#include <net/if_var.h>
#include <netinet/if_ether.h>
#include <net/ether.h>

static HANDLE arpExThread = 0;

/*!
 * \addtogroup xgARP
 */
/*@{*/

/*! \fn slArpExpire(void *arg)
 * \brief ARP expiration thread.
 *
 * Running once a minute to handle expired 
 * ARP cache entries.
 */
THREAD(slArpExpire, arg)
{
    ARPENTRY * volatile ae;
    ARPENTRY * volatile aer = 0;
    ARPENTRY ** volatile aep;
    NUTDEVICE *dev;
    IFNET *ifn;

    for(;;) {
        slSleep(60000);

        /*
         * Loop through the list of registered devices
         * and process all net devices.
         */
        for(dev = nutDeviceList; dev; dev = dev->dev_next) {
            if(dev->dev_type == IFTYP_NET) {
                ifn = dev->dev_icb;
                ae = ifn->arpTable;
                aep = &ifn->arpTable;

                /*
                 * Loop through the list of ARP entries and
                 * remove outdated ones.
                 */
                while(ae) {
                    if(ae->ae_outdated++ >= 9) {
                        *aep = ae->ae_next;
                        aer = ae;
                    }
                    if(aer == 0)
                        aep = &ae->ae_next;
                    ae = ae->ae_next;
                    if(aer) {
                        slHeapFree(aer);
                        aer = 0;
                    }
                }
            }
        }
    }
}


/*!
 * \brief Locate ARP entry of a specified interface by a specified IP address.
 *
 * \note A read lock must have been set before calling this function.
 *
 * \param ifn Pointer to the interface.
 * \param ip  IP address.
 *
 * \return Pointer to the ARP table entry, if found, or NULL
 *         if no entry exists.
 */
static ARPENTRY * volatile slArpCacheLookup(IFNET *ifn, u_long ip)
{
    ARPENTRY * volatile entry;

    for(entry = ifn->arpTable; entry; entry = entry->ae_next) {
        if(entry->ae_ip == ip)
            break;
    }
    return entry;
}


/*!
 * \brief Create a new ARP cache entry.
 *
 * \note A read lock must have been set before calling this function.
 *
 * If the cache is locked by another reader, the entry is not added.
 * Later we may add a backlog to queue completed entries.
 */
static ARPENTRY *slArpCacheNew(IFNET *ifn, u_long ip, u_char *ha, u_char flags)
{
    u_char rc = 0;
    ARPENTRY *entry = 0;

    if((entry = slHeapAllocClear(sizeof(ARPENTRY))) != 0) {
        entry->ae_ip = ip;
        if(ha)
            memcpy(entry->ae_ha, ha, 6);
        entry->ae_flags = flags;

        slEnterCritical();
        entry->ae_next = ifn->arpTable;
        ifn->arpTable = entry;
        rc = 1;
        slExitCritical();

        if(rc == 0) {
            slHeapFree(entry);
            entry = 0;
        }
    }
    return entry;
}

/*!
 * \brief Update an ARP entry.
 *
 * If an entry with the same IP address exists, then this entry is 
 * updated. If no entry exists, a new one is created. All threads 
 * waiting for address resolution are woken up.
 *
 * \note This function is automatically called on each incoming
 *       ARP telegram. Applications typically do not call this 
 *       function.
 *
 * \param dev Identifies the device, which cache is updated.
 *            This pointer must have been retrieved previously
 *            by using the slDeviceOpen() function.
 * \param ip  Requested IP address in network byte order.
 * \param ha  Pointer to a buffer which receives the MAC address.
 */
void slArpCacheUpdate(NUTDEVICE *dev, u_long ip, u_char *ha)
{
    ARPENTRY *entry;

    if(arpExThread == 0)
        arpExThread = slThreadCreate("arpex", slArpExpire, NULL, 384);

    entry = slArpCacheLookup(dev->dev_icb, ip);

    /*
     * Entry found, update it.
     */
    if(entry) {

        /*
         * Do not update permanent entries.
         */
        if((entry->ae_flags & ATF_PERM) == 0) {
            entry->ae_outdated = 0;
            memcpy(entry->ae_ha, ha, 6);
            entry->ae_flags |= ATF_COM;
        }

        /*
         * Wake all waiting threads.
         */
        slEventBroadcast(&entry->ae_tq);
        return;
    }

    /*
     * Entry not found. Try to add a new one.
     */
    slArpCacheNew(dev->dev_icb, ip, ha, ATF_COM);
}

/*!
 * \brief Query MAC address for a specified IP address.
 *
 * If no entry is available in the ARP cache, an ARP request
 * is generated, sent out to the network and the calling
 * thread will be suspended.
 *
 * \param dev  Identifies the device to be used for sending.
 *             This pointer must have been retrieved previously
 *             by using the slDeviceOpen() function.
 * \param ip   IP address of which the caller asked the MAC address.
 * \param mac  Buffer for the retrieved MAC address.
 *
 * \return 0 if address resolved, -1 otherwise.
 */
int slArpCacheQuery(NUTDEVICE *dev, CONST u_long ip, u_char *mac)
{
    volatile ARPENTRY *entry;
    NETBUF *nb = 0;
    int rc = -1;
    u_char retries = 4;

    /*
     * Search a matching entry. If we found a completed
     * one, return the MAC address to the caller.
     */
    if((entry = slArpCacheLookup(dev->dev_icb, ip)) != 0) {
        if(entry->ae_flags & ATF_COM) {
            memcpy(mac, ((ARPENTRY *)entry)->ae_ha, 6);
            return 0;
        }
    }

    /*
     * No entry exists. Create a new incomplete entry.
     * This may fail if we ran out of memeory.
     */
    else if((entry = slArpCacheNew(dev->dev_icb, ip, 0, 0)) == 0) {
        return -1;
    }

    /*
     * We have created a new entry or found an incomplete one.
     * Allocate a request packet.
     */
    if((nb = slArpAllocNetBuf(ARPOP_REQUEST, ip, 0)) == 0) {
        return -1;
    }

    /*
     * We enter a loop, which will send ARP requests on fixed
     * time intervals until our ARP entry gets completed.
     */
    while(retries--) {
        if(slArpOutput(dev, nb)) 
            break;
        if(entry->ae_flags & ATF_COM) {
            memcpy(mac, ((ARPENTRY *)entry)->ae_ha, 6);
            rc = 0;
            break;
        }

        /*
         * Sleep until woken up by a new ARP
         * response from the net or on timeout.
         */
        slEventWait(&((ARPENTRY *)entry)->ae_tq, 500);
    }
    slNetBufFree(nb);
    return rc;
}

/*@}*/
