
#include <net/if_var.h>
#include <netinet/if_ether.h>

/*!
 * \brief Handle incoming ARP packets.
 *
 * Packets not destined to us or packets with unsupported 
 * address type or item length are silently discarded.
 *
 * \note This routine is called by the Ethernet layer on
 *       incoming ARP packets. Applications typically do
 *       not call this function.
 *
 * \param dev Identifies the device that received the packet.
 * \param nb  Pointer to a network buffer structure containing 
 *            the ARP packet.
 */
void slArpInput(NUTDEVICE *dev, NETBUF *nb)
{
    ETHERARP *ea;
    ARPHDR *ah;
    IFNET *nif;

    ea = (ETHERARP *)nb->nb_nw.vp;
    ah = (ARPHDR *)&ea->ea_hdr;

    /*
     * Silently discard packets with unsupported
     * address types and lengths.
     */
    if(ntohs(ah->ar_hrd) != ARPHRD_ETHER ||
       ntohs(ah->ar_pro) != ETHERTYPE_IP ||
       ah->ar_hln != 6 ||
       ah->ar_pln != 4) {
        slNetBufFree(nb);
        return;
    }

    /*
     * Silently discard packets for other destinations.
     */
    nif = dev->dev_icb;
    if(ea->arp_tpa != nif->if_local_ip) {
        slNetBufFree(nb);
        return;
    }

    /*
     * TODO: Silently discard packets with our own 
     * source address.
     */

    /*
     * TODO: Discard packets with broadcast source
     * address.
     */

    /*
     * TODO: Discard packets if source IP address
     * equals our address. We should send a reply.
     */

    /*
     * Add the sender to our arp cache. Note, that we don't do
     * this with replies only, but also with requests on our
     * address. The assumption is that if someone is requesting
     * our address, they are probably intending to talk to us,
     * so it saves time if we cache their address.
     */
    slArpCacheUpdate(dev, ea->arp_spa, ea->arp_sha);

    /*
     * Reply to ARP requests.
     */
    if(htons(ea->ea_hdr.ar_op) == ARPOP_REQUEST) {
        NETBUF *nbr = slArpAllocNetBuf(ARPOP_REPLY, ea->arp_spa, ea->arp_sha);

        if(nbr) {
            slArpOutput(dev, nbr);
            slNetBufFree(nbr);
        }
    }
    slNetBufFree(nb);
}
/*@}*/
