
#include <string.h>

#include <sys/heap.h>
#include <sys/thread.h>

#include <net/if_var.h>
#include <netinet/if_ether.h>
#include <net/ether.h>

/*!
 * \brief Allocate an ARP network buffer structure.
 *
 * \param type Type of ARP packet.
 * \param ip   Target IP address.
 * \param mac  Target MAC address, null pointer for broadcast.
 *
 * \return Pointer to the allocated network buffer structure 
 *         or 0 on failure.
 */
NETBUF *slArpAllocNetBuf(u_short type, u_long ip, u_char *mac)
{
    NETBUF *nb;
    ETHERARP *ea;
    ARPHDR *ah;

    if((nb = slNetBufAlloc(0, NBAF_NETWORK, sizeof(ETHERARP))) == 0)
        return 0;

    ea = nb->nb_nw.vp;
    ah = (ARPHDR *)&ea->ea_hdr;

    /*
     * Set ARP header.
     */
    ah->ar_hrd = htons(ARPHRD_ETHER);
    ah->ar_pro = htons(ETHERTYPE_IP);
    ah->ar_hln = 6;
    ah->ar_pln = 4;
    ah->ar_op = htons(type);

    /*
     * Set ARP destination data.
     */
    if(mac)
        memcpy(ea->arp_tha, mac, 6);
    else
        memset(ea->arp_tha, 0xff, 6);
    ea->arp_tpa = ip;

    return nb;
}

/*!
 * \brief Send an ARP packet.
 *
 * \note Applications typically do not call this function.
 *
 * \param dev   Identifies the device to use.
 * \param nb    Network buffer structure containing the packet to be sent.
 *              The structure must have been allocated by a previous
 *              call slNetBufAlloc().
 *
 * \return 0 on success, -1 in case of any errors.
 */
int slArpOutput(NUTDEVICE *dev, NETBUF *nb)
{
    ETHERARP *ea;
    IFNET *nif;

    ea = nb->nb_nw.vp;

    /*
     * Set ARP source data.
     */
    nif = dev->dev_icb;
    memcpy(ea->arp_sha, nif->if_mac, 6);
    ea->arp_spa = nif->if_local_ip;

    return slEtherOutput(dev, ETHERTYPE_ARP, ea->arp_tha, nb);
}

/*@}*/
