
#include <netinet/in.h>
#include <netinet/ip_icmp.h>
#include <netinet/ipcsum.h>
#include <netinet/icmp.h>


/*!
 * \addtogroup xgICMP
 */
/*@{*/
/*!
 * \brief Send an ICMP datagram.
 *
 * Known ICMP types are:
 *
 * - #ICMP_ECHOREPLY Echo reply
 * - #ICMP_UNREACH Destination unreachable
 * - #ICMP_SOURCEQUENCH Packet lost
 * - #ICMP_REDIRECT Shorter route
 * - #ICMP_ECHO Echo reply
 * - #ICMP_ROUTERADVERT Router advertisement
 * - #ICMP_ROUTERSOLICIT Router solicitation
 * - #ICMP_TIMXCEED Time exceeded
 * - #ICMP_PARAMPROB Bad IP header
 * - #ICMP_TSTAMP Timestamp request
 * - #ICMP_TSTAMPREPLY Timestamp reply
 * - #ICMP_IREQ Information request
 * - #ICMP_IREQREPLY Information reply
 * - #ICMP_MASKREQ Address mask request
 * - #ICMP_MASKREPLY Address mask reply
 *
 * \param type Type of the ICMP message.
 * \param dest Destination IP address.
 * \param nb   Network buffer structure containing the datagram.
 *
 * \return 0 on success, -1 otherwise.
 */
int slIcmpOutput(u_char type, u_long dest, NETBUF *nb)
{
    ICMPHDR *icp;

    icp = (ICMPHDR *)nb->nb_tp.vp;
    icp->icmp_type = type;
    icp->icmp_cksum = 0;
    icp->icmp_cksum = slIpChkSum(0, nb->nb_tp.vp, nb->nb_tp.sz);

    return slIpOutput(IPPROTO_ICMP, dest, nb);
}

/*!
 * \brief Send an ICMP message to a given destination.
 *
 * \param type Type of the ICMP message. See slIcmpOutput() for 
 *             a list of valid types.
 * \param code Type subcode.
 * \param spec Type specific data item.
 * \param dest IP address of the target.
 * \param nb   Network buffer structure of this message. 
 *             The application part must have been set by 
 *             the caller and should not use this buffer
 *             again, because it's released on return.
 *
 * \return 0 on success, -1 otherwise.
 */
int slIcmpReply(u_char type, u_char code, u_long spec, u_long dest, NETBUF *nb)
{
    ICMPHDR *icp;
    u_short csum;

    nb = slNetBufAlloc(nb, NBAF_TRANSPORT, sizeof(ICMPHDR));

    icp = (ICMPHDR *)nb->nb_tp.vp;
    icp->icmp_type = type;
    icp->icmp_code = code;
    icp->icmp_spec = spec;
    icp->icmp_cksum = 0;

    csum = slIpChkSumPartial(0, nb->nb_tp.vp, nb->nb_tp.sz);
    icp->icmp_cksum = slIpChkSum(csum, nb->nb_ap.vp, nb->nb_ap.sz);

    return slIpOutput(IPPROTO_ICMP, dest, nb);
}

/*@}*/
