
#include <string.h>

#include <eeprom.h>

#include <net/ether.h>
#include <net/route.h>
#include <arpa/inet.h>

#include <sys/confnet.h>

#include <pro/dhcp.h>


CONFNET confnet;

static u_char egnite_null_mac[] = { 0x12,0x34,0x56,0x00,0x00,0x01 };

/*!
 * \brief Load network configuration from EEPROM.
 *
 * \param name Name of the device.
 *
 * \return 0 if configuration has been read. Otherwise the
 *         return value is -1, IP address is 0, IP netmask
 *         is 255.255.255.0 and MAC is 00-06-98-00-00-00.
 */
int slNetLoadConfig(CONST char *name)
{
    eeprom_read_block(&confnet, CONFNET_EE_OFFSET, sizeof(CONFNET));
    if(confnet.cd_size == sizeof(CONFNET) && strcmp(confnet.cd_name, name) == 0) 
        return 0;

    memcpy(confnet.cdn_mac, egnite_null_mac, 6);
    confnet.cdn_ip_addr = 0;
    confnet.cdn_ip_mask = inet_addr("255.255.255.0");
    confnet.cdn_cip_addr = 0;
    confnet.cdn_gateway = 0;
    return -1;
}

/*!
 * Save network configuration in EEPROM.
 */
int slNetSaveConfig(void)
{
    u_char *cp;
    int i;

    confnet.cd_size = sizeof(CONFNET);
    for(cp = (u_char *)&confnet, i = 0; i < sizeof(CONFNET); i++) 
        eeprom_wb(i + CONFNET_EE_OFFSET, *cp++);

    return 0;
}

/*!
 * Network interface setup.
 *
 * \param dev  Identifies the network device to setup. This pointer must 
 *             have been retrieved previously by calling slDeviceOpen().
 * \param ip_addr
 * \param ip_mask
 * \param gateway
 *
 * \return 0 on success, -1 otherwise.
 *
 * \note Applications typically do not call this function.
 *       It's automatically called by slNetIfConfig() or
 *       indirectly by slNetAutoConfig().
 */
int slNetIfSetup(NUTDEVICE *dev, u_long ip_addr, u_long ip_mask, u_long gateway)
{
    IFNET *nif;

    nif = dev->dev_icb;

    /*
     * Use specified or previously used IP address.
     */
    if(ip_addr == 0 && (ip_addr = confnet.cdn_ip_addr) == 0)
        return -1;
    nif->if_local_ip = ip_addr;

    /*
     * Use specified or default mask.
     */
    if(ip_mask == 0)
        ip_mask = inet_addr("255.255.255.0");
    nif->if_mask = ip_mask;

    /*
     * Add routing entries.
     */
    slIpRouteAdd(ip_addr & ip_mask, ip_mask, 0, dev);
    if(gateway)
        slIpRouteAdd(0, 0, gateway, dev);

    /*
     * Save configuration in EEPROM.
     */
    memcpy(confnet.cd_name, dev->dev_name, sizeof(confnet.cd_name));
    memcpy(confnet.cdn_mac, nif->if_mac, sizeof(nif->if_mac));
    confnet.cdn_ip_addr = ip_addr;
    confnet.cdn_ip_mask = ip_mask;

    /*
     * Set gateway, if one was provided by the caller. Remove
     * gateway, if it's outside of our network.
     */
    if(gateway || (confnet.cdn_gateway & ip_mask) != (ip_addr & ip_mask))
        confnet.cdn_gateway = gateway;

    return slNetSaveConfig();
}

/*!
 * \brief Configure a network interface.
 *
 * Open the device and associate the given MAC address, IP address and 
 * netmask with the network interface.
 *
 * \param name    Name of the device to configure. This device must have
 *                been registered by calling slRegisterDevice().
 * \param mac     6-byte Ethernet MAC address to be used. This must be
 *                a unique address within the local Ethernet. The vendor
 *                of the Ethernut board will supply an address, which is
 *                unique within the known universe.
 * \param ip_addr IP address to be used in network byte order. This must 
 *                be a unique address within the Internet. If you do not 
 *                directly communicate with other Internet hosts, you can 
 *                use a locally assigned address.
 * \param ip_mask IP network mask to be used in network byte order. 
 *                Typical networks use 255.255.255.0, which allows upto 
 *                254 hosts.
 *
 * \return 0 on success, -1 otherwise.
 */
int slNetIfConfig(CONST char *name, CONST u_char *mac, u_long ip_addr, u_long ip_mask)
{
    NUTDEVICE *dev;
    IFNET *nif;

    /*
     * Open device by name.
     */
    if((dev = slDeviceOpen(name)) == 0) 
        return -1;

    /*
     * We support Ethernet only.
     */
    if(dev->dev_type != IFTYP_NET)
        return -1;
    nif = dev->dev_icb;
    if(nif->if_type != IFT_ETHER)
        return -1;
    nif->if_recv = slEtherInput;

    /*
     * Initialize interface.
     */
    memcpy(nif->if_mac, mac, sizeof(nif->if_mac));
    if((*dev->dev_init)(dev)) 
        return -1;

    /*
     * If we got a valid IP address, create 
     * a route and return.
     */
    if(ip_addr & ip_mask) {
        slNetIfSetup(dev, ip_addr, ip_mask, 0);
        return 0;
    }

    /*
     * Start DHCP client and wait for any result.
     */
    return slDhcpClientStart(dev);
}


/*!
 * \brief Configure a network interface using EEPROM values.
 *
 * Loads IP address, netmask and MAC address from EEPROM and calls
 * slNetIfConfig().
 *
 * \param name Name of the device.
 *
 * \return 0 on success, -1 otherwise.
 */
int slNetAutoConfig(CONST char *name)
{
    slNetLoadConfig(name);
    return slNetIfConfig(name, confnet.cdn_mac, confnet.cdn_cip_addr, confnet.cdn_ip_mask);
}

/*@}*/
