
#include <arpa/inet.h>

/*!
 * \brief Convert decimal dotted ASCII representation into numeric IP address.
 *
 * \param str String containing the ASCII representation.
 *
 * \return IP address in network byte order.
 */
u_long inet_addr(CONST u_char *str)
{
    u_short num;
    u_long addr = 0;
    u_char parts = 0;
    u_char *ap;

    ap = (u_char *)&addr;
    while(parts < 4) {
        if(*str < '0' || *str > '9')
            break;
        for(num = 0; num <= 255;) {
            num = (num * 10) + (*str - '0');
            if(*++str < '0' || *str > '9')
                break;
        }
        if(num > 255)
            break;
        parts++;
        *ap++ = (u_char)num;
        if(*str != '.') {
            if(parts == 4)
                return addr;
            break;
        }
        str++;
    }
    return -1;
}

/*!
 * \brief Convert numeric IP address into decimal dotted 
 *        ASCII representation.
 *
 * \note This function is not thread safe. Each subsequent
 *       call will destroy the previous result. Applications
 *       should locally store the result before calling the 
 *       function again or allowing other threads to call it.
 *
 * \param addr IP address in network byte order.
 *
 * \return Pointer to a static buffer containing the 
 *         ASCII representation.
 */
u_char *inet_ntoa(u_long addr)
{
    static u_char str[16];
    u_char inv[3];
    u_char *rp;
    u_char *ap;
    u_char rem;
    u_char n;
    u_char i;

    rp = str;
    ap = (u_char *)&addr;
    for(n = 0; n < 4; n++) {
        i = 0;
        do {
            rem = *ap % (u_char)10;
            *ap /= (u_char)10;
            inv[i++] = '0' + rem;
        } while(*ap);
        while(i--)
            *rp++ = inv[i];
        *rp++ = '.';
        ap++;
    }
    *--rp = 0;
    return str;
}
/*@}*/
