
#include <netinet/ipcsum.h>


/*!
 * \brief Calculates a partial IP checksum over a block of data.
 *
 * Note that this returns the checksum in network byte order, and thus does
 * not need to be converted via hton16(), etc.  Of course this means that
 * we mustn't use this value for normal arithmetic!
 *
 * This is a partial checksum because it doesn't take the 1's complement
 * of the overall sum.
 *
 * \note by Dave Hudson: This doesn't look particularly intuitive at 
 * first sight - in fact it probably looks plain wrong.  It does work 
 * however (take my word for it), but for some explanation the reader 
 * is referred to RFC1071 where the maths is explained in detail.
 *
 * \note by Harald Kipp: Yes, the GCC version looks wrong. I assume that
 *       this will not work for all packet sizes. Not yet confirmed.
 */
u_short slIpChkSumPartial(u_short partial_csum, void *buf, u_short count)
{
#ifdef __GNUC__
    u_short words;
    void *d1;
    u_short d2;
	d1 = buf;
    words = count >> 1;
    if(words != 0) {
        asm volatile(
            "\n"
            "L_redo%=:\n\t"
            "ld __tmp_reg__, %a1+\n\t"
            "add %A0, __tmp_reg__\n\t"
            "ld __tmp_reg__, %a1+\n\t"
            "adc %B0, __tmp_reg__\n\t"
            "adc %A0, r1\n\t"
            "adc %B0, r1\n\t"
            "dec %A2\n\t"
            "brne L_redo%=\n\t"
            "subi %B2, 1\n\t"
            "brsh L_redo%=\n\t"
            "\n\t"
            : "=r" (partial_csum), "=e" (d1), "=w" (d2)
            : "0" (partial_csum), "1" (d1), "2" (words)
        );
    }

    /*
     * Did we have an odd number of bytes to do?
     */
    if(count & 0x01) {
        asm volatile (
            "ld __tmp_reg__, %a1+\n\t"
            "add %A0, __tmp_reg__\n\t"
            "adc %B0, r1\n\t"
            "adc %A0, r1\n\t"
            : "=r" (partial_csum), "=e" (d1)
            : "0" (partial_csum), "1" (d1)
        );      
    }
    
#else
//*KU* implementation of checksum
// has to be optimized later !

/*	u_long sum = partial_csum;
	u_short *d2 = (u_short*) buf;
	while (count > 1) 
	{
		sum += *d2++;	
		count -= 2;
	}
	if (count > 0)
		sum += *(u_char*) d2++;
		
	while (sum>>16)
		sum = (sum & 0xFFFF) + (sum >>16);

	partial_csum = sum;		*/


/*KU* this should be much faster */
    	asm("ldd r24, Y+0");		// pop count from stack
    	asm("ldd r25, Y+1");
        asm("clr r1");			// zero reg
    	asm("clr r8");			// registers for short sum
    	asm("clr r9");
    	asm("clr r4");			// registers for long sum
    	asm("clr r5");
    	asm("clr r6");
    	asm("clr r7");   	
    	asm("add r4, r16");	        // initialize sum with partial_csum
    	asm("adc r5, r17");
    	asm("movw r26, r24");   	// check count if zero
    	asm("subi r26, 0");		
    	asm("sbci r27, 0");	
    	asm("movw r8, r16");	    	// returnvalue if exit here
    	asm("breq slIPChk_Exit"); 	
    	asm("subi r26, 1");	        // check count = 1
    	asm("sbci r27, 0");
    	asm("movw r30, r18");		// load buffer address
    	asm("breq slIPChk_OneByte");
    		
	asm("slIPChk_Start:");
	asm("ld r8, Z+");		// load value from buf
    	asm("ld r9, Z+");    	
    	asm("add r4, r8");		// add to long sum
    	asm("adc r5, r9");
    	asm("adc r6, r1");
    	asm("adc r7, r1");
    	asm("subi r24, 2");		// count - 2; next word
    	asm("sbci r25, 0");
    	asm("breq slIPChk_Fold");
    	asm("movw r26, r24");		// check count = 1
    	asm("subi r26, 1");
    	asm("sbci r27, 0");
    	asm("breq slIPChk_OneByte");
    	asm("rjmp slIPChk_Start");	// there are more words
    	
    	asm("slIPChk_OneByte:");	// no word, only one byte left
    	asm("ld r8, Z");		// load last byte
    	asm("add r4, r8");		// add it to long sum
    	asm("adc r5, r1");
    	asm("adc r6, r1");
    	asm("adc r7, r1");
    	
    	asm("slIPChk_Fold:");		// see the while loop in the
    	asm("movw r8, r4");		// source code above if you
    	asm("add r8, r6");		// want to understand this
    	asm("adc r9, r7");
    	asm("brcc slIPChk_Exit"); 	// fold again if sum>>16 = true
    	asm("ldi r26, 1");
    	asm("add r8, r26");	
    	asm("adc r9, r1");
    	
    	asm("slIPChk_Exit:");
    	asm("movw %partial_csum, r8");	// pass the return value
    	    	
#endif

    return partial_csum; 
}
         
    	

/*!
 * \brief Calculates an the final IP checksum over a block of data.
 *
 * Unlike the partial checksum in slIpChkSumPartial(), this function takes
 * the one's complement of the final result, thus making it the full checksum.
 */
u_short slIpChkSum(u_short partial_csum, void *buf, u_short count)
{
    return slIpChkSumPartial(partial_csum, buf, count) ^ 0xffff;
}

/*
 * Details of the pseudo header used as part of the
 * calculation of UDP and TCP header checksums.
 */
struct pseudo_hdr {
    u_long ph_src_addr;
    u_long ph_dest_addr;
    u_char ph_zero;
    u_char ph_protocol;
    u_short ph_len;
};

/*!
 * \brief Calculates the partial IP pseudo checksum.
 *
 */
u_short slIpPseudoChkSumPartial(u_long src_addr, u_long dest_addr, u_char protocol, u_short len)
{
    struct pseudo_hdr ph;

    ph.ph_src_addr = src_addr;
    ph.ph_dest_addr = dest_addr;
    ph.ph_zero = 0;
    ph.ph_protocol = protocol;
    ph.ph_len = len;

    return slIpChkSumPartial(0, &ph, sizeof(struct pseudo_hdr));
}

/*!
 * \brief Calculates the IP pseudo checksum.
 *
 */
u_short slIpPseudoChkSum(u_long src_addr, u_long dest_addr, u_char protocol, u_short len)
{
    return slIpPseudoChkSumPartial(src_addr, dest_addr, protocol, len) ^ 0xffff;
}

/*@}*/
