
#include <net/route.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/icmp.h>
#include <netinet/udp.h>
#include <sys/socket.h>
#include <arpa/inet.h>

/*!
 * \brief Process incoming IP datagrams.
 *
 * Datagrams addressed to other destinations and datagrams 
 * whose version number is not 4 are silently discarded.
 *
 * \note This routine is called by the Ethernet layer on
 *       incoming IP datagrams. Applications typically do
 *       not call this function.
 *
 * \param dev Identifies the device that received this datagram.
 * \param nb  The network buffer received.
 */
void slIpInput(NUTDEVICE *dev, NETBUF *nb)
{
    IPHDR *ip;
    u_short ip_hdrlen;
    u_long dst;
    u_char bcast;
    IFNET *nif;

    ip = nb->nb_nw.vp;

    /*
     * Silently discard datagrams of different IP version
     * and fragmented datagrams.
     */
    if(ip->ip_v != IPVERSION || (ntohs(ip->ip_off) & (IP_MF | IP_OFFMASK)) != 0) {
        slNetBufFree(nb);
        return;
    }

    ip_hdrlen = ip->ip_hl * 4;

    /*
     * No checksum calculation on incoming datagrams!
     */

    /*
     * Check for broadcast.
     */
    dst = ip->ip_dst;
    nif = dev->dev_icb;

    if(dst == 0xffffffff ||
        (nif->if_local_ip && (dst | nif->if_mask) == 0xffffffff)) {
        bcast = 1;
    }
    else {
        bcast = 0;

        if(nif->if_local_ip == 0 && 
           ip->ip_p == IPPROTO_ICMP &&
           (dst & 0xff000000) != 0xff000000 &&
           (dst & 0xff000000) != 0) {
            slNetIfSetup(dev, dst, 0, 0);
        }

        /*
         * Silently discard datagrams for other destinations.
         * However, if we haven't got an IP address yet, we
         * allow ICMP datagrams to support dynamic IP ARP method.
         */
        if(dst == 0 || dst != nif->if_local_ip) {
            slNetBufFree(nb);
            return;
        }
    }

    nb->nb_nw.sz = ip_hdrlen;
    nb->nb_tp.vp = ((u_char *)ip) + (ip_hdrlen);
    nb->nb_tp.sz = htons(ip->ip_len) - (ip_hdrlen);

    switch(ip->ip_p) {
    case IPPROTO_ICMP:
        slIcmpInput(dev, nb);
        break;
    case IPPROTO_UDP:
        slUdpInput(nb);
        break;
    case IPPROTO_TCP:
        /*
         * Silently discard TCP broadcasts.
         */
        if(bcast)
            slNetBufFree(nb);
        else 
            slTcpInput(nb);
        break;
    default:
        slNetBufFree(nb);
        break;
    }
}

/*@}*/
