
#include <string.h>

#include <netinet/if_ether.h>
#include <net/ether.h>
#include <netinet/ipcsum.h>
#include <net/route.h>
#include <netinet/ip.h>


/*!
 * \brief Send IP datagram.
 *
 * Route an IP datagram to the proper interface.
 *
 * The function will not return until the data has been stored 
 * in the network device hardware for transmission. If the 
 * device is not ready for transmitting a new packet, the 
 * calling thread will be suspended until the device becomes 
 * ready again. If the hardware address of the target host needs 
 * to be resolved the function will be suspended too.
 *
 * \param proto Protocol type.
 * \param dest  Destination IP address. The function will determine
 *              the proper network interface by checking the routing
 *              table. It will also perform any neccessary hardware
 *              address resolution.
 * \param nb    Network buffer structure containing the datagram. 
 *
 * \return 0 on success, -1 otherwise.
 *
 * \bug Broadcasts to multiple network devices are currently not supported.
 */
int slIpOutput(u_char proto, u_long dest, NETBUF *nb)
{
    u_char ha[6];
    IPHDR *ip;
    NUTDEVICE *dev;
    IFNET *nif;
    u_long gate;

    if((nb = slNetBufAlloc(nb, NBAF_NETWORK, sizeof(IPHDR))) == 0)
        return -1;

    ip = nb->nb_nw.vp;
    ip->ip_v = 4;
    ip->ip_hl = sizeof(IPHDR) / 4;
    ip->ip_tos = 0;
    ip->ip_len = htons(nb->nb_nw.sz + nb->nb_tp.sz + nb->nb_ap.sz);
    ip->ip_off = 0;
    ip->ip_ttl = 0x40;
    ip->ip_p = proto;
    ip->ip_dst = dest;

    /*
     * Broadcasts are directly sent to all interfaces.
     */
    if(dest == 0xffffffff) {

        memset(ha, 0xff, sizeof(ha));

        for(dev = nutDeviceList; dev; dev = dev->dev_next) {
            if(dev->dev_type == IFTYP_NET) {
                nif = dev->dev_icb;
                ip->ip_id = htons(nif->if_pkt_id++);
                ip->ip_src = nif->if_local_ip;
                ip->ip_sum = 0;
                ip->ip_sum = slIpChkSum(0, nb->nb_nw.vp, nb->nb_nw.sz);

                slEtherOutput(dev, ETHERTYPE_IP, ha, nb);
            }
        }
        return 0;
    }

    /*
     * Get destination's route.
     * This will also return the proper interface.
     */
    if((dev = slIpRouteQuery(dest, &gate)) == 0) 
        return -1;

    nif = dev->dev_icb;
    ip->ip_id = htons(nif->if_pkt_id++);
    ip->ip_src = nif->if_local_ip;
    ip->ip_sum = 0;
    ip->ip_sum = slIpChkSum(0, nb->nb_nw.vp, nb->nb_nw.sz);

    /*
     * Query the MAC address of our destination. If the destination is
     * not a member of the local network, we query the MAC address of
     * the gateway.
     */
    if(slArpCacheQuery(dev, gate ? gate : dest, ha)) {
        return -1;
    }
    return slEtherOutput(dev, ETHERTYPE_IP, ha, nb);
}

/*@}*/
