
#include <sys/atom.h>
#include <sys/heap.h>

#include <net/if_var.h>
#include <net/route.h>


static u_char rteLock;

RTENTRY *rteList = 0;     //!< Linked list of routing entries.

/*!
 * \brief Add a new entry to the IP routing table.
 *
 * Note, that there is currently no support for
 * removing entries or detecting duplicates.
 * Anyway, newer entries will be found first,
 * because they are inserted in front of older
 * entries. However, this works only for equal
 * masks, i.e. new network routes will never
 * overwrite old host routes.
 *
 * \param ip   Network or host IP address to be route.
 *             Set 0 for default route.
 * \param mask Mask for this entry. -1 for host routes,
 *             0 for default or net mask for net routes.
 * \param gate Route through this gateway, otherwise 0.
 * \param dev  Network interface to use.
 *
 * \return 0 on success, -1 otherwise.
 */
int slIpRouteAdd(u_long ip, u_long mask, u_long gate, NUTDEVICE *dev)
{
    int rc = -1;
    RTENTRY *rte;
    RTENTRY *rtp;
    RTENTRY **rtpp;

    AtomicInc(&rteLock);
    slEnterCritical();
    if(rteLock == 1) {

        /*
         * Insert a new entry into the list, which is
         * sorted based on the mask. Host routes are
         * in front, default routes at the end and
         * networks in between.
         */
        rtp = rteList;
        rtpp = &rteList;
        while(rtp && rtp->rt_mask > mask) {
            rtpp = &rtp->rt_next;
            rtp = rtp->rt_next;
        }
        if((rte = slHeapAlloc(sizeof(RTENTRY))) != 0) {
            rte->rt_ip = ip & mask;
            rte->rt_mask = mask;
            rte->rt_gateway = gate;
            rte->rt_dev = dev;
            rte->rt_next = rtp;
            *rtpp = rte;
            rc = 0;
        }
    }
    slExitCritical();
    AtomicDec(&rteLock);

    return rc;
}

/*!
 * \brief Used to limit route lookup recursion for gateways.
 *
 * The returned pointer points directly into the linked
 * list of all route entries, which is save, as entries
 * are never removed or modified.
 *
 * \param ip    IP address in network byte order.
 * \param gate  Points to a buffer which optionally
 *              receives the IP address of a gateway.
 *              The pointer may be NULL, if the caller
 *              is not interested in this information.
 * \param level Recursion level.
 */
static RTENTRY *slIpRouteRecQuery(u_long ip, u_long *gate, u_char level)
{
    RTENTRY *rte = 0;

    if(level < 4) {
        for(rte = rteList; rte; rte = rte->rt_next) {
            if((ip & rte->rt_mask) == rte->rt_ip)
                break;
        }
        if(rte) {
            if(gate && rte->rt_gateway)
                *gate = rte->rt_gateway;
            if(rte->rt_dev == 0)
                rte = slIpRouteRecQuery(rte->rt_gateway, gate, level + 1);
        }
    }
    return rte;
}

/*!
 * \brief Find a device associated with a particular IP route.
 *
 * Gateway routes will be automatically resolved up to
 * four levels of redirection.
 *
 * \param ip    IP address to find a route for, given in
 *              network byte order.
 * \param gate  Points to a buffer which optionally
 *              receives the IP address of a gateway.
 *              The pointer may be NULL, if the caller
 *              is not interested in this information.
 *
 * \return Pointer to the interface structure or NULL
 *         if no route was found.
 */
NUTDEVICE *slIpRouteQuery(u_long ip, u_long *gate)
{
    RTENTRY *rte;

    if(gate)
        *gate = 0;
    AtomicInc(&rteLock);
    rte = slIpRouteRecQuery(ip, gate, 0);
    AtomicDec(&rteLock);

    return rte ? rte->rt_dev : 0;
}

/*@}*/
