
#include <sys/heap.h>
#include <netinet/sostream.h>

/*!
 * \brief Wait for input.
 *
 * This function checks the input buffer for any data. If
 * the buffer is empty, the calling \ref xrThread "thread" 
 * will be blocked until new data arrives. 
 *
 * \param dev Identifies the stream device, which has been created 
 *            previously by calling slSoStreamCreate().
 *
 * \return 0 on success, -1 otherwise.
 */
static int slSoStreamInput(NUTDEVICE *dev)
{
    IFSTREAM *ifs;
    TCPSOCKET *sock;
    int got;

    ifs = dev->dev_icb;
    ifs->if_rd_idx = 0;
    sock = dev->dev_dcb;
    if((got = slTcpReceive(sock, ifs->if_rx_buf, sizeof(ifs->if_rx_buf))) <= 0) {
        ifs->if_rx_idx = 0;
        return got;
    }
    ifs->if_rx_idx = (u_char)got;
    return 0;
}

/*!
 * \brief Initiate output.
 *
 * This function checks the output buffer. If it contains
 * any data, a data segment is transmitted to the remote 
 * host.
 *
 * \param dev Identifies the stream device, which has been created 
 *            previously by calling slSoStreamCreate().
 *
 * \return 0 on success, -1 otherwise.
 */
static int slSoStreamOutput(NUTDEVICE *dev)
{
    IFSTREAM *ifs;
    TCPSOCKET *sock;
    u_char n;
    int c;

    ifs = dev->dev_icb;
    if(ifs->if_wr_idx) {
        sock = dev->dev_dcb;
        for(n = 0; n < ifs->if_wr_idx; n += c) 
            if((c = slTcpSend(sock, ifs->if_tx_buf + n, ifs->if_wr_idx - n)) < 0)
                return -1;
        ifs->if_wr_idx = 0;
    }
    return 0;
}

/*!
 * \brief Wait for output buffer empty.
 *
 * If the output buffer contains any data, the calling
 * thread is suspended until all data has been transmitted.
 *
 * \param dev Identifies the stream device, which has been created 
 *            previously by calling slSoStreamCreate().
 */
static int slSoStreamFlush(NUTDEVICE *dev)
{
    return slSoStreamOutput(dev);
}

/*!
 * \brief Create a virtual stream device from a specified socket.
 *
 * \param sock Specifies the socket descriptor, which must have been 
 *             retrieved by calling slTcpCreateSocket(). In
 *             addition a connection must already have been established
 *             by calling slTcpConnect() or slTcpAccept().
 */
NUTDEVICE *slSoStreamCreate(TCPSOCKET *sock)
{
    NUTDEVICE *dev;
    IFSTREAM *ifs;

    if((dev = slHeapAllocClear(sizeof(NUTDEVICE))) != 0) {
        dev->dev_type = IFTYP_STREAM;
        dev->dev_dcb = sock;
        if((dev->dev_icb = slHeapAllocClear(sizeof(IFSTREAM))) != 0) {
            ifs = dev->dev_icb;
            ifs->if_input = slSoStreamInput;
            ifs->if_output = slSoStreamOutput;
            ifs->if_flush = slSoStreamFlush;
        }
        else {
            slHeapFree(dev);
            dev = 0;
        }
    }

    return dev;
}

/*!
 * \brief Destroy a previously created stream device.
 *
 * This will flush the stream output buffer and release any
 * occupied memory. To terminate the connection and close
 * the socket, the application must call slTcpCloseSocket().
 *
 * \param dev Identifies the stream device, which has been created 
 *            previously by calling slSoStreamCreate().
 */
int slSoStreamDestroy(NUTDEVICE *dev)
{
    slSoStreamFlush(dev);
    slHeapFree(dev->dev_icb);
    slHeapFree(dev);

    return 0;
}


/*@}*/
