
#include <string.h>

#include <sys/thread.h>
#include <sys/heap.h>
#include <sys/event.h>
#include <sys/atom.h>

#include <net/errno.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/icmp.h>
#include <netinet/ip_icmp.h>
#include <netinet/ipcsum.h>
#include <net/route.h>
#include <sys/socket.h>
#include <netinet/tcputil.h>
#include <netinet/tcp.h>


extern TCPSOCKET * volatile tcpSocketList;

/*!
 * \addtogroup xgTCP
 */
/*@{*/

HANDLE tcp_in_rdy;
NETBUF * volatile tcp_in_nbq;
volatile u_short tcp_in_cnt;
static HANDLE tcpThread = 0;

/* ================================================================
 * Helper functions
 * ================================================================
 */

/*
 * Nothing done yet.
 */
static void slTcpInputOptions(TCPSOCKET *sock, NETBUF *nb)
{
}

/*!
 * \brief Move application data from the network buffer structure 
 *        to the socket's receive buffer.
 *
 * \param sock Socket descriptor. This pointer must have been 
 *             retrieved by calling slTcpCreateSocket().
 * \param nb   Network buffer structure containing a TCP segment.
 */
static void slTcpProcessAppData(TCPSOCKET *sock, NETBUF *nb)
{
    u_short total;

    if((total = nb->nb_ap.sz + sock->so_rx_cnt) < sock->so_rx_bsz) {
        if(sock->so_rx_buf == 0) {
            if(slHeapAvailable() > sock->so_rx_bsz + 2048)
                sock->so_rx_buf = slHeapAlloc(sock->so_rx_bsz);
        }
        if(sock->so_rx_buf) {
            memcpy(sock->so_rx_buf + sock->so_rx_cnt, nb->nb_ap.vp, nb->nb_ap.sz);
            sock->so_rx_cnt = total;

            sock->so_rx_nxt += nb->nb_ap.sz;
            if(nb->nb_ap.sz >= sock->so_rx_win)
                sock->so_rx_win = 0;
            else
                sock->so_rx_win -= nb->nb_ap.sz;
            if(nb->nb_next == 0 && tcp_in_nbq == 0)
                sock->so_tx_flags |= SO_FORCE | SO_ACK;
            else
                sock->so_tx_flags |= SO_ACK;

            /*
             * Wake up a thread waiting for data.
             */
            slEventPost(&sock->so_rx_tq);

            slTcpOutput(sock, 0, 0);
        }
    }
}

/*
 * \param sock Socket descriptor.
 */
static void slTcpProcessSyn(TCPSOCKET *sock, IPHDR *ih, TCPHDR *th)
{
    u_short mss;
    NUTDEVICE *dev;
    IFNET *nif;

    sock->so_local_addr = ih->ip_dst;
    sock->so_remote_port = th->th_sport;
    sock->so_remote_addr = ih->ip_src;

    sock->so_rx_nxt =
    sock->so_tx_wl1 =
    sock->so_rx_isn = ntohl(th->th_seq);
    sock->so_rx_nxt++;
    sock->so_tx_win = ntohs(th->th_win);

    /*
     * To avoid unneccessary fragmentation, limit the
     * maximum segment size to the maximum transfer
     * unit of our interface.
     */
    if((dev = slIpRouteQuery(ih->ip_src, 0)) != 0) {
        nif = dev->dev_icb;
        mss = nif->if_mtu - sizeof(IPHDR) - sizeof(TCPHDR);
        if(sock->so_mss == 0 || sock->so_mss > mss)
            sock->so_mss = mss;
    }
}

/*!
 * \brief ACK processing.
 *
 * \param sock Socket descriptor. This pointer must have been 
 *             retrieved by calling slTcpCreateSocket().
 *
 */
static void slTcpProcessAck(TCPSOCKET *sock, TCPHDR *th, u_short length)
{
    NETBUF *nb;
    u_long h_seq;
    u_long h_ack;

    /*
     * If remote acked something not yet send, reply immediately.
     */
    h_ack = ntohl(th->th_ack);
    if(h_ack > sock->so_tx_nxt) {
        sock->so_tx_flags |= SO_ACK | SO_FORCE;
        return;
    }

    /*
     * If the new sequence number or acknowledged sequence number
     * is above our last update, we adjust our transmit window.
     * Avoid dupe ACK processing on window updates.
     */
    h_seq = ntohl(th->th_seq);
    if(h_ack == sock->so_tx_una) {
        if(h_seq > sock->so_tx_wl1 || (h_seq == sock->so_tx_wl1 && h_ack >= sock->so_tx_wl2)) {
            sock->so_tx_win = ntohs(th->th_win);
            sock->so_tx_wl1 = h_seq;
            sock->so_tx_wl2 = h_ack;
        }
    }

    /*
     * Ignore old ACKs but wake up sleeping transmitter threads, because
     * the window size may have changed.
     */
    if(h_ack < sock->so_tx_una) {
        slEventPost(&sock->so_tx_tq);
        return;
    }

    /*
     * Process duplicate ACKs.
     */
    if(h_ack == sock->so_tx_una) {
        /*
         * Don't count, if nothing is waiting for ACK,
         * segment contains data or on SYN/FIN segments.
         */
        if(sock->so_tx_nbq && length == 0 && (th->th_flags & (TH_SYN | TH_FIN)) == 0) {
            /*
             * If dupe counter reaches it's limit, resend
             * the oldest unacknowledged netbuf.
             */
            if(++sock->so_tx_dup >= 3) {
                sock->so_retran_time = 0;
                sock->so_tx_dup = 0;
#ifdef NUTDEBUG
                if(tcp_trace)
                    slKDumpTcpHeader("RET", sock, sock->so_tx_nbq);
#endif
                slIpOutput(IPPROTO_TCP, sock->so_remote_addr, sock->so_tx_nbq);
            }
        }
        slEventPost(&sock->so_tx_tq);
        return;
    }

    /*
     * We're here, so the ACK must have actually acked something
     */
    sock->so_tx_dup = 0;
    sock->so_tx_una = h_ack;

    /*
     * Remove all acknowledged netbufs.
     */
    while((nb = sock->so_tx_nbq) != 0) {
        if(ntohl(((TCPHDR *)(nb->nb_tp.vp))->th_seq) + nb->nb_ap.sz > h_ack)
            break;
        sock->so_tx_nbq = nb->nb_next;
        slNetBufFree(nb);
    }
    
    /*
     * Reset retransmit timer and wake up waiting transmissions.
     */
    sock->so_retran_time = 0;
    slEventPost(&sock->so_tx_tq);
}



/* ================================================================
 * State changes.
 * ================================================================
 */
/*!
 * State change, possibly inform application.
 *
 * \param sock  Socket descriptor.
 * \param state New state to switch to.
 *
 * \return 0 on success, -1 on illegal state change attempt.
 */
static int slTcpStateChange(TCPSOCKET *sock, u_char state)
{
    int rc = 0;

    switch(sock->so_state) {
    case TCPS_LISTEN:
        /*
         * SYN received.
         */
        if(state == TCPS_SYN_RECEIVED) {
            sock->so_tx_flags |= SO_SYN | SO_ACK;
            slTcpOutput(sock, 0, 0);
        }
        else
            rc = -1;
        break;

    case TCPS_SYN_SENT:
        switch(state) {
        case TCPS_LISTEN:
            /*
             * RST received on passive socket.
             */
            break;
        case TCPS_SYN_RECEIVED:
            /*
             * SYN received.
             */
            sock->so_tx_flags |= SO_SYN | SO_ACK;
            slTcpOutput(sock, 0, 0);
            break;
        case TCPS_ESTABLISHED:
            /*
             * SYNACK received.
             */
            sock->so_tx_flags |= SO_ACK | SO_FORCE;
            slTcpOutput(sock, 0, 0);
            break;
        default:
            rc = -1;
            break;
        }
        break;

    case TCPS_SYN_RECEIVED:
        switch(state) {
        case TCPS_LISTEN:
            /*
             * RST received on passive socket.
             */
            break;
        case TCPS_ESTABLISHED:
            /*
             * ACK of SYN received.
             */
            /* wake up accepting thread on passive socket. */
            slEventPost(&sock->so_pc_tq);
            break;
        case TCPS_FIN_WAIT_1:
            /*
             * Closed by application.
             */
            sock->so_tx_flags |= SO_FIN;
            slTcpOutput(sock, 0, 0);
            break;
        case TCPS_CLOSE_WAIT:
            /*
             * FIN received.
             */
            sock->so_tx_flags |= SO_FIN | SO_ACK;
            slTcpOutput(sock, 0, 0);
            break;
        default:
            rc = -1;
            break;
        }
        break;

    case TCPS_ESTABLISHED:
        switch(state) {
        case TCPS_FIN_WAIT_1:
            /*
             * Closed by application.
             */
            sock->so_tx_flags |= SO_FIN | SO_ACK;
            slTcpOutput(sock, 0, 0);
            break;
        case TCPS_CLOSE_WAIT:
            /*
             * FIN received.
             */
            sock->so_tx_flags |= SO_ACK | SO_FORCE;
            slTcpOutput(sock, 0, 0);
            /*
             * Inform application.
             */
            slEventBroadcast(&sock->so_rx_tq);
            slEventBroadcast(&sock->so_tx_tq);
            slEventBroadcast(&sock->so_pc_tq);
            slEventBroadcast(&sock->so_ac_tq);
            break;
        default:
            rc = -1;
            break;
        }
        break;

    case TCPS_FIN_WAIT_1:
        switch(state) {
        case TCPS_FIN_WAIT_2:
            /*
             * ACK of FIN received.
             */
            break;
        case TCPS_CLOSING:
            /*
             * FIN received.
             */
            break;
        case TCPS_TIME_WAIT:
            /*
             * FIN and ACK of FIN received.
             */
            break;
        default:
            rc = -1;
            break;
        }
        break;

    case TCPS_FIN_WAIT_2:
        /*
         * FIN received.
         */
        if(state != TCPS_TIME_WAIT)
            rc = -1;
        sock->so_tx_flags |= SO_ACK | SO_FORCE;
        slTcpOutput(sock, 0, 0);
        break;

    case TCPS_CLOSE_WAIT:
        /*
         * Closed by application.
         */
        if(state == TCPS_LAST_ACK) {
            sock->so_tx_flags |= SO_FIN | SO_ACK;
            slTcpOutput(sock, 0, 0);
        }
        else
            rc = -1;
        break;

    case TCPS_CLOSING:
        /*
         * ACK of FIN received.
         */
        if(state != TCPS_TIME_WAIT)
            rc = -1;
        break;

    case TCPS_LAST_ACK:
        rc = -1;
        break;

    case TCPS_TIME_WAIT:
        rc = -1;
        break;

    case TCPS_CLOSED:
        switch(state) {
        case TCPS_LISTEN:
            /*
             * Passive open by application.
             */
            break;
        case TCPS_SYN_SENT:
            /*
             * Active open by application.
             */
            sock->so_tx_flags |= SO_SYN;
            if(slTcpOutput(sock, 0, 0)) {
                rc = -1;
                sock->so_last_error = EHOSTDOWN;
                slEventPost(&sock->so_ac_tq);
            }
            break;
        default:
            rc = -1;
            break;
        }
        break;
    }
#ifdef NUTDEBUG
    if(tcp_trace) {
        slPrintFormat(0, " %04x-", sock);
        if(rc)
            slKDumpSockState(sock->so_state, "**ERR ", "**>");
        slKDumpSockState(state, "[>", "]");
    }
#endif
    if(rc == 0)
        sock->so_state = state;
    return rc;
}

/* ================================================================
 * Application events.
 * ================================================================
 */
/*!
 * \brief Initiated by the application.
 *
 * \param sock Socket descriptor. This pointer must have been 
 *             retrieved by calling slTcpCreateSocket().
 *
 * \return 0 if granted, error code otherwise.
 */
int slTcpStatePassiveOpenEvent(TCPSOCKET *sock)
{
    if(sock->so_state != TCPS_CLOSED)
        return (sock->so_last_error = EISCONN);

    slTcpStateChange(sock, TCPS_LISTEN);

    /*
     * Block application.
     */
    slEventWait(&sock->so_pc_tq, 0);

    return 0;
}

/*!
 * \brief Initiated by the application.
 *
 * Optionally specify the local IP address.
 *
 * \param sock Socket descriptor. This pointer must have been 
 *             retrieved by calling slTcpCreateSocket().
 *
 * \return 0 if granted, -1 otherwise.
 */
int slTcpStateActiveOpenEvent(TCPSOCKET *sock)
{
    /*
     * Switch state to SYN_SENT. This will
     * transmit a SYN packet.
     */
    slTcpStateChange(sock, TCPS_SYN_SENT);

    /*
     * Block application.
     */
    slEventWait(&sock->so_ac_tq, 0);

    if(sock->so_state != TCPS_ESTABLISHED)
        return -1;
    return 0;
}

/*!
 * \brief Socket close by application.
 *
 * If socket is in state SYN_RECEIVED or ESTABLISHED,
 * it is changed to FINWAIT1.
 *
 * No further data sending is accepted.
 * Receiving is still allowed.
 *
 * \param sock Socket descriptor. This pointer must have been 
 *             retrieved by calling slTcpCreateSocket().
 */
int slTcpStateCloseEvent(TCPSOCKET *sock)
{
    if(sock == 0)
        return -1;

    switch(sock->so_state) {
    case TCPS_LISTEN:
    case TCPS_SYN_SENT:
        /*
         * No connection yet, immediately destroy the socket.
         */
        slTcpDestroySocket(sock);
        break;

    case TCPS_SYN_RECEIVED:
    case TCPS_ESTABLISHED:
        /*
         * Send a FIN and wait for ACK or FIN.
         */
        slTcpStateChange(sock, TCPS_FIN_WAIT_1);
        break;

    case TCPS_CLOSE_WAIT:
        /* 
         * RFC 793 is wrong. 
         */
        slTcpStateChange(sock, TCPS_LAST_ACK);
        break;

    case TCPS_FIN_WAIT_1:
    case TCPS_FIN_WAIT_2:
    case TCPS_CLOSING:
    case TCPS_LAST_ACK:
    case TCPS_TIME_WAIT:
        sock->so_last_error = EALREADY;
        return -1;

    default:
        sock->so_last_error = ENOTCONN;
        return -1;
    }
    return 0;
}

/*!
 * \brief Initiated by the application.
 *
 * \param sock Socket descriptor. This pointer must have been 
 *             retrieved by calling slTcpCreateSocket().
 */
int slTcpStateWindowEvent(TCPSOCKET *sock)
{
    if(sock == 0)
        return -1;
    sock->so_tx_flags |= SO_ACK | SO_FORCE;
    slTcpOutput(sock, 0, 0);

    return 0;
}

/* ================================================================
 * Timeout events.
 * ================================================================
 */
/*!
 * \brief Retransmit a segment after ACK timeout.
 *
 * This function is called by the TCP timer.
 *
 * \param sock Socket descriptor. This pointer must have been 
 *             retrieved by calling slTcpCreateSocket().
 */
void slTcpStateRetranTimeout(TCPSOCKET *sock)
{
    if(sock->so_retran_time > 300) {
        sock->so_time_wait = 0;
        sock->so_state = TCPS_CLOSE_WAIT;
        slEventBroadcast(&sock->so_ac_tq);
        slEventBroadcast(&sock->so_rx_tq);
        slEventBroadcast(&sock->so_tx_tq);
        slEventBroadcast(&sock->so_pc_tq);
    }
    else {
#ifdef NUTDEBUG
        if(tcp_trace)
            slKDumpTcpHeader("RET", sock, sock->so_tx_nbq);
#endif
        if(slIpOutput(IPPROTO_TCP, sock->so_remote_addr, sock->so_tx_nbq)) {
            sock->so_tx_nbq = sock->so_tx_nbq->nb_next;
            sock->so_time_wait = 0;
            sock->so_state = TCPS_CLOSE_WAIT;
            slEventBroadcast(&sock->so_ac_tq);
            slEventBroadcast(&sock->so_rx_tq);
            slEventBroadcast(&sock->so_tx_tq);
            slEventBroadcast(&sock->so_pc_tq);
        }
    }
}

/* ================================================================
 * Segment arrival events.
 * ================================================================
 */

/*!
 * \brief
 * Process incoming segments in listening state.
 *
 * Wait for a connection request from a remote socket.
 *
 * \param sock Socket descriptor.
 * \param nb   Network buffer structure containing a TCP segment.
 */
static void slTcpStateListen(TCPSOCKET *sock, u_char flags, TCPHDR *th, NETBUF *nb)
{
    /*
     * Got a SYN segment. Store relevant data in our socket
     * structure and switch to TCPS_SYN_RECEIVED.
     */
    if((flags & (TH_SYN | TH_ACK | TH_RST)) == TH_SYN) {
        slTcpProcessSyn(sock, nb->nb_nw.vp, th);
        slTcpStateChange(sock, TCPS_SYN_RECEIVED);
        slNetBufFree(nb);
    }
    else
        slTcpReject(nb);
}


/*!
 * \brief Process incoming segments in SYN-SENT state.
 *
 * Wait for a matching connection request after having sent ours.
 *
 * \param sock Socket descriptor.
 * \param nb   Network buffer structure containing a TCP segment.
 */
static void slTcpStateSynSent(TCPSOCKET *sock, u_char flags, TCPHDR *th, NETBUF *nb)
{
    /*
     * Validate ACK, if set.
     */
    if(flags & TH_ACK) {
        if(!IsInLimits(ntohl(th->th_ack), sock->so_tx_isn + 1, sock->so_tx_nxt)) {
            slTcpReject(nb);
            return;
        }
    }

    /*
     * Handle RST flag. If we were in the LISTEN state,
     * then we return to the LISTEN state, otherwise we
     * abort the connection and go to the CLOSED state.
     */
    if(flags & TH_RST) {
        if(flags & TH_ACK) {
            if(sock->so_pc_tq)
                slTcpStateChange(sock, TCPS_LISTEN);
            else {
                sock->so_last_error = ECONNREFUSED;
                sock->so_state = TCPS_CLOSE_WAIT;
            }
        }
        slNetBufFree(nb);
        return;
    }

    /*
     * Handle SYN flag. If we got a valid ACK too, then
     * our connection is established. Otherwise enter
     * SYNRCVD state.
     */
    if(flags & TH_SYN) {
        slTcpProcessSyn(sock, nb->nb_nw.vp, th);
        if(flags & TH_ACK) {
            slTcpProcessAck(sock, th, nb->nb_ap.sz);
            slTcpStateChange(sock, TCPS_ESTABLISHED);
            slEventPost(&sock->so_ac_tq);
        }
        else {
            slTcpStateChange(sock, TCPS_SYN_RECEIVED);
        }
    }
    slNetBufFree(nb);
}

/*
 * \brief
 * Process incoming segments in SYN-RECEIVED state.
 *
 * Waiting for a confirming connection request
 * acknowledgment after having both received
 * and sent a connection request.
 *
 * \param sock Socket descriptor.
 * \param nb   Network buffer structure containing a TCP segment.
 */
static void slTcpStateSynReceived(TCPSOCKET *sock, u_char flags, TCPHDR *th, NETBUF *nb)
{
    /*
     * If our previous ack receives a reset response,
     * then we fall back to the listening state.
     */
    if(flags & TH_RST) {
        slTcpStateChange(sock, TCPS_LISTEN);
        slNetBufFree(nb);
        return;
    }

    /*
     * Silently discard segments without ACK.
     */
    if((flags & TH_ACK) == 0) {
        slNetBufFree(nb);
        return;
    }

    /*
     * Reject out of window sequence.
     */
    if(!IsInLimits(ntohl(th->th_ack), sock->so_tx_una + 1, sock->so_tx_nxt)) {
        slTcpReject(nb);
        return;
    }

    /*
     * Reject SYNs.
     */
    if(flags & TH_SYN) {
        slTcpReject(nb);
        return;
    }

    /*
     * Process acknowledge and application data
     * and release the network buffer.
     */
    slTcpProcessAck(sock, th, nb->nb_ap.sz);
    if(nb->nb_ap.sz)
        slTcpProcessAppData(sock, nb);
    slNetBufFree(nb);

    /*
     * Process state change.
     */
    if(flags & TH_FIN) {
        sock->so_rx_nxt++;
        slTcpStateChange(sock, TCPS_CLOSE_WAIT);
    }
    else
        slTcpStateChange(sock, TCPS_ESTABLISHED);
}

/*
 * \brief Process incoming segments from established connections.
 *
 * Received application data will be delivered to the application
 * until we receive a FIN segment.
 *
 * \param sock  Socket descriptor.
 * \param flags TCP flags.
 * \param th    Pointer to the TCP header within the NETBUF.
 * \param nb    Network buffer structure containing a TCP segment.
 */
static void slTcpStateEstablished(TCPSOCKET *sock, u_char flags, TCPHDR *th, NETBUF *nb)
{
    if(flags & TH_RST) {
        sock->so_state = TCPS_CLOSE_WAIT;
        slNetBufFree(nb);
        slEventBroadcast(&sock->so_rx_tq);
        slEventBroadcast(&sock->so_tx_tq);
        slEventBroadcast(&sock->so_pc_tq);
        slEventBroadcast(&sock->so_ac_tq);
        return;
    }

    /*
     * Silently discard segments without ACK.
     */
    if((flags & TH_ACK) == 0) {
        slNetBufFree(nb);
        return;
    }

    /*
     * Reject SYNs. Silently discard late SYNs
     * (Thanks to Mike Cornelius).
     */
    if(flags & TH_SYN) {
        if(htonl(th->th_seq) != sock->so_rx_isn)
            slTcpReject(nb);
        return;
    }

    slTcpProcessAck(sock, th, nb->nb_ap.sz);

    if(htonl(th->th_seq) > sock->so_rx_nxt) {
        NETBUF *nbq;
        NETBUF **nbqp;
        TCPHDR *thq;

        if(nb->nb_ap.sz) {
            nbq = sock->so_rx_nbq;
            nbqp = &sock->so_rx_nbq;
            while(nbq) {
                thq = (TCPHDR *)(nbq->nb_tp.vp);
                if(htonl(th->th_seq) < htonl(thq->th_seq)) {
                    *nbqp = nb;
                    nb->nb_next = nbq;
                    break;
                }
                if(htonl(th->th_seq) == htonl(thq->th_seq)) {
                    slNetBufFree(nb);
                    sock->so_tx_flags |= SO_ACK | SO_FORCE;
                    slTcpOutput(sock, 0, 0);
                    return;
                }
                nbqp = &nbq->nb_next;
                nbq = nbq->nb_next;
            }
            if(nbq == 0) {
                *nbqp = nb;
                nb->nb_next = 0;
            }
        }
        else {
            slNetBufFree(nb);
        }
        sock->so_tx_flags |= SO_ACK | SO_FORCE;
        slTcpOutput(sock, 0, 0);
        return;
    }

    /*
     * Acknowledge any sequence numbers not expected, 
     * even if they do not contain any data. Keepalive
     * packets contain a sequence number one less
     * than the next data expected and they do not 
     * contain any data.
     */
    if(htonl(th->th_seq) != sock->so_rx_nxt) {
        sock->so_tx_flags |= SO_ACK | SO_FORCE;
        sock->so_oos_drop++;
        slNetBufFree(nb);
        slTcpOutput(sock, 0, 0);
    }
    else if(nb->nb_ap.sz) {
        slTcpProcessAppData(sock, nb);
        slNetBufFree(nb);
        while((nb = sock->so_rx_nbq) != 0) {
            th = (TCPHDR *)(nb->nb_tp.vp);
            if(htonl(th->th_seq) > sock->so_rx_nxt) 
                break;
            sock->so_rx_nbq = nb->nb_next;
            if(htonl(th->th_seq) >= sock->so_rx_nxt) {
                slTcpProcessAppData(sock, nb);
                flags |= th->th_flags;
            }
            slNetBufFree(nb);
        }
    }
    else
        slNetBufFree(nb);

    if(flags & TH_FIN) {
        sock->so_rx_nxt++;
        slTcpStateChange(sock, TCPS_CLOSE_WAIT);
    }
}

/*
 * \brief Process incoming segments in FIN-WAIT1 state.
 *
 * Waiting for a connection termination request
 * from the remote, or an acknowledgment of the
 * connection termination request previously sent.
 *
 * The application already closed the socket.
 *
 * \param sock Socket descriptor.
 * \param nb   Network buffer structure containing a TCP segment.
 */
static void slTcpStateFinWait1(TCPSOCKET *sock, u_char flags, TCPHDR *th, NETBUF *nb)
{
    if(flags & TH_RST) {
        slNetBufFree(nb);
        slTcpDestroySocket(sock);
        return;
    }

    /*
     * Silently discard segments without ACK.
     */
    if((flags & TH_ACK) == 0) {
        slNetBufFree(nb);
        return;
    }

    /*
     * Reject SYNs.
     */
    if(flags & TH_SYN) {
        slTcpReject(nb);
        return;
    }

    slTcpProcessAck(sock, th, nb->nb_ap.sz);
    slTcpStateChange(sock, TCPS_FIN_WAIT_2);

    /*
     * Process application data and release the network buffer.
     */
    if(nb->nb_ap.sz)
        slTcpProcessAppData(sock, nb);
    slNetBufFree(nb);

    if(flags & TH_FIN) {
        sock->so_rx_nxt++;
        /* 
         * Our FIN has been acked.
         */
        sock->so_time_wait = 0;
        slTcpStateChange(sock, TCPS_TIME_WAIT);
    }
}

/*
 * \brief Process incoming segments in FIN-WAIT2 state.
 *
 * Waiting for a connection termination request
 * from the remote.
 *
 * The application already closed the socket.
 *
 * \param sock Socket descriptor.
 * \param nb   Network buffer structure containing a TCP segment.
 */
static void slTcpStateFinWait2(TCPSOCKET *sock, u_char flags, TCPHDR *th, NETBUF *nb)
{
    if(flags & TH_RST) {
        slNetBufFree(nb);
        slTcpDestroySocket(sock);
        return;
    }

    /*
     * Silently discard segments without ACK.
     */
    if((flags & TH_ACK) == 0) {
        slNetBufFree(nb);
        return;
    }

    /*
     * Reject SYNs.
     */
    if(flags & TH_SYN) {
        slTcpReject(nb);
        return;
    }

    /*
     * Process acknowledge and application data and release the 
     * network buffer.
     */
    slTcpProcessAck(sock, th, nb->nb_ap.sz);
    if(nb->nb_ap.sz)
        slTcpProcessAppData(sock, nb);
    slNetBufFree(nb);

    if(flags & TH_FIN) {
        sock->so_rx_nxt++;
        sock->so_time_wait = 0;
        slTcpStateChange(sock, TCPS_TIME_WAIT);
    }
}

/*
 * \brief
 * Process incoming segments in CLOSE-WAIT state.
 *
 * Waiting for a connection termination request
 * from the local application.
 *
 * \param sock Socket descriptor.
 * \param nb   Network buffer structure containing a TCP segment.
 */
static void slTcpStateCloseWait(TCPSOCKET *sock, u_char flags, TCPHDR *th, NETBUF *nb)
{
    if(flags & TH_RST) {
        slNetBufFree(nb);
        return;
    }

    /*
     * Silently discard segments without ACK.
     */
    if((flags & TH_ACK) == 0) {
        slNetBufFree(nb);
        return;
    }

    /*
     * Reject SYNs.
     */
    if(flags & TH_SYN) {
        slTcpReject(nb);
        return;
    }

    slTcpProcessAck(sock, th, nb->nb_ap.sz);

    slNetBufFree(nb);
}

/*
 * \brief
 * Process incoming segments in CLOSING state.
 *
 * Waiting for a connection termination request
 * acknowledgment from the remote.
 *
 * The application already closed the socket.
 *
 * \param sock Socket descriptor.
 * \param nb   Network buffer structure containing a TCP segment.
 */
static void slTcpStateClosing(TCPSOCKET *sock, u_char flags, TCPHDR *th, NETBUF *nb)
{
    if(flags & TH_RST) {
        slNetBufFree(nb);
        slTcpDestroySocket(sock);
        return;
    }

    /*
     * Silently discard segments without ACK.
     */
    if((flags & TH_ACK) == 0) {
        slNetBufFree(nb);
        return;
    }

    /*
     * Reject SYNs.
     */
    if(flags & TH_SYN) {
        slTcpReject(nb);
        return;
    }

    slTcpProcessAck(sock, th, nb->nb_ap.sz);

    sock->so_time_wait = 0;
    slTcpStateChange(sock, TCPS_TIME_WAIT);

    slNetBufFree(nb);
}

/*!
 * \brief
 * Process incoming segment in LAST_ACK state.
 *
 * Waiting for an acknowledgment of the connection termination
 * request previously sent.
 *
 * The application already closed the socket.
 *
 * \param sock  Socket descriptor.
 * \param flags TCP flags of incoming segment.
 * \param th    TCP header of incoming segment.
 * \param nb    Network buffer structure containing a TCP segment.
 */
static void slTcpStateLastAck(TCPSOCKET *sock, u_char flags, TCPHDR *th, NETBUF *nb)
{
    if(flags & TH_RST) {
        slNetBufFree(nb);
        slTcpDestroySocket(sock);
        return;
    }

    /*
     * Silently discard segments without ACK.
     */
    if((flags & TH_ACK) == 0) {
        slNetBufFree(nb);
        return;
    }

    /*
     * Reject SYNs.
     */
    if(flags & TH_SYN) {
        slTcpReject(nb);
        return;
    }

    slTcpProcessAck(sock, th, nb->nb_ap.sz);

    slNetBufFree(nb);

    slTcpDestroySocket(sock);
}

/*!
 * \brief Process incoming TCP segments.
 *
 * Processing is based on the current state of the socket connection.
 *
 * \param sock Socket descriptor. This pointer must have been 
 *             retrieved by calling slTcpCreateSocket().
 * \param nb   Network buffer structure containing a TCP segment.
 *             Will be released within this routine.
 */
static void slTcpStateProcess(TCPSOCKET *sock, NETBUF *nb)
{
    TCPHDR *th = (TCPHDR *)nb->nb_tp.vp;
    u_char flags = th->th_flags;

#ifdef NUTDEBUG
    if(tcp_trace) {
        slPrintFormat(0, " %04x-", sock);
        slKDumpSockState(sock->so_state, "[", ">]");
    }
#endif
    switch(sock->so_state) {
    case TCPS_LISTEN:
        slTcpStateListen(sock, flags, th, nb);
        break;
    case TCPS_SYN_SENT:
        slTcpStateSynSent(sock, flags, th, nb);
        break;
    case TCPS_SYN_RECEIVED:
        slTcpStateSynReceived(sock, flags, th, nb);
        break;
    case TCPS_ESTABLISHED:
        slTcpStateEstablished(sock, flags, th, nb);
        break;
    case TCPS_FIN_WAIT_1:
        slTcpStateFinWait1(sock, flags, th, nb);
        break;
    case TCPS_FIN_WAIT_2:
        slTcpStateFinWait2(sock, flags, th, nb);
        break;
    case TCPS_CLOSE_WAIT:
        slTcpStateCloseWait(sock, flags, th, nb);
        break;
    case TCPS_CLOSING:
        slTcpStateClosing(sock, flags, th, nb);
        break;
    case TCPS_LAST_ACK:
        slTcpStateLastAck(sock, flags, th, nb);
        break;
    case TCPS_TIME_WAIT:
        /*
         * Ignore everything while in TIME_WAIT state.
         */
        slNetBufFree(nb);
        break;
    case TCPS_CLOSED:
        /*
         * Reject everything while in CLOSED state.
         */
        slTcpReject(nb);
        break;
    default:
        slNetBufFree(nb);
        break;
    }
}

/*! \fn NicTx(void *arg)
 * \brief TCP state machine thread.
 *
 */
THREAD(slTcpSm, arg)
{
    NETBUF *nb;
    NETBUF *nbx;
    TCPHDR *th;
    IPHDR *ih;
    TCPSOCKET *sock;

    /*
     * Give us a very high priority to avoid queuing too many packets.
     */
    //slThreadSetPriority(9);

    for(;;) {
        if(slEventWait(&tcp_in_rdy, 200)) {
            for(sock = tcpSocketList; sock; sock = sock->so_next) {

                /*
                 * Send late acks.
                 */
                if(sock->so_tx_flags & SO_ACK) {
                    sock->so_tx_flags |= SO_FORCE;
                    slTcpOutput(sock, 0, 0);
                }

                /*
                 * Process retransmit timer.
                 */
                if(sock->so_tx_nbq) {
                    if(sock->so_state == TCPS_ESTABLISHED) {
                        /*
                         * For slow connections you may have to replace the 
                         * 3 by 15. Thanks to Mike Cornelius.
                         *
                         * Yes, we really need round trip time calculation.
                         */
                        if((++(sock->so_retran_time) & 3) == 0) {
                            slTcpStateRetranTimeout(sock);
                        }
                    }
                    else if((++(sock->so_retran_time) & 15) == 0) {
                        slTcpStateRetranTimeout(sock);
                    }
                }

                /*
                 * Destroy sockets after timeout in TIMEWAIT state.
                 */
                if(sock->so_state == TCPS_TIME_WAIT || 
                   sock->so_state == TCPS_FIN_WAIT_1 ||
                   sock->so_state == TCPS_FIN_WAIT_2) {
                    if(sock->so_time_wait++ >= 9) {
                        slTcpDestroySocket(sock);
                        break;
                    }
                }

                /*
                 * Recover from SYN flood attacks.
                 */
                else if(sock->so_state == TCPS_SYN_RECEIVED) {
                    if(sock->so_time_wait++ >= 45) {
                        sock->so_state = TCPS_LISTEN;
                        sock->so_time_wait = 0;
                    }
                }
            }
        }
        else {
            nb = tcp_in_nbq;
            tcp_in_nbq = 0;
            tcp_in_cnt = 0;
            while(nb) {
                ih = (IPHDR *)nb->nb_nw.vp;
                th = (TCPHDR *)nb->nb_tp.vp;
                sock = slTcpFindSocket(th->th_dport, th->th_sport, ih->ip_src);
#ifdef NUTDEBUG
                if(tcp_trace)
                    slKDumpTcpHeader(" IN", sock, nb);
#endif
                nbx = nb->nb_next;
                if(sock) {
                    slTcpInputOptions(sock, nb);
                    slTcpStateProcess(sock, nb);
                }

                /*
                 * Reject the segment, if no matching socket was found.
                 */
                else 
                    slTcpReject(nb);
                nb = nbx;
            }
        }
    }
}

/*!
 * \brief Process incoming TCP segments.
 */
void slTcpStateMachine(NETBUF *nb)
{
    NETBUF *nbp;
    u_short size;

    nb->nb_next = 0;

    if((nbp = tcp_in_nbq) == 0) {
        tcp_in_nbq = nb;
        slEventPost(&tcp_in_rdy);
    }
    else {
        size = nb->nb_nw.sz + nb->nb_tp.sz + nb->nb_ap.sz;
        if(tcp_in_cnt + size + 2048 < slHeapAvailable()) {
            tcp_in_cnt += size;
            while(nbp->nb_next)
                nbp = nbp->nb_next;
            nbp->nb_next = nb;
            slEventPost(&tcp_in_rdy);
        }
        else 
            slNetBufFree(nb);
    }
}


int slTcpInitStateMachine(void)
{
    if(tcpThread == 0 && 
       (tcpThread = slThreadCreate("tcpsm", slTcpSm, NULL, 1024)) == 0)
        return -1;
    return 0;
}

/*@}*/
