
#include <string.h>

#include <sys/heap.h>
#include <sys/event.h>

#include <netinet/ip.h>
#include <netinet/udp.h>
#include <sys/socket.h>


UDPSOCKET *udpSocketList = 0;    /*!< Linked list of all UDP sockets. */
static u_short last_local_port = 4096;  /* Unassigned local port. */

/*!
 * \brief Create an UDP socket.
 *
 * \param  port Server applications provide the local port number
 *              with this parameter. Client applications should
 *              pass zero.
 *
 * \return Socket descriptor of the newly created UDP socket or
 *         0 if there is not enough memory left.
 *
 */
UDPSOCKET *slUdpCreateSocket(u_short port)
{
    UDPSOCKET *sock;

    if(port == 0) {
        do {
            if(++last_local_port == 0)
                last_local_port = 4096;

            sock = udpSocketList;
            while(sock) {
                if(sock->so_local_port == last_local_port)
                    break;
                sock = sock->so_next;
            }
        } while(sock);
        port = last_local_port;
    }
    if((sock = slHeapAllocClear(sizeof(UDPSOCKET))) != 0) {
        sock->so_local_port = htons(port);
        sock->so_next = udpSocketList;
        udpSocketList = sock;
    }
    return sock;
}

/*!
 * \brief Send an UDP datagram.
 *
 * \param sock Socket descriptor. This pointer must have been 
 *             retrieved by calling slUdpCreateSocket().
 * \param addr IP address of the remote host in network byte order.
 * \param port Remote port number in host byte order.
 * \param data Pointer to a buffer containing the data to send.
 * \param len  Number of bytes to be sent.
 *
 * \return 0 on success, -1 otherwise.
 */
int slUdpSendTo(UDPSOCKET *sock, u_long addr, u_short port, void *data, u_short len)
{
    int rc;
    NETBUF *nb;

    if((nb = slNetBufAlloc(0, NBAF_APPLICATION, len)) == 0) {
        return -1;
    }
    memcpy(nb->nb_ap.vp, data, len);

    rc = slUdpOutput(sock, addr, port, nb);
    slNetBufFree(nb);

    return rc;
}

/*!
 * \brief Receive an UDP datagram.
 *
 * \param sock    Socket descriptor. This pointer must have been 
 *                retrieved by calling slUdpCreateSocket().
 * \param addr    IP address of the remote host in network byte order.
 * \param port    Remote port number in host byte order.
 * \param data    Pointer to the buffer that receives the data.
 * \param size    Size of the buffer that receives the data.
 * \param timeout Maximum number of milliseconds to wait.
 *
 * \return The number of bytes received, if successful. The return
 *         value -1 indicates an error. On timeout 0 is returned.
 *
 * \note Timeout is limited to the granularity of the system timer.
 */
int slUdpReceiveFrom(UDPSOCKET *sock, u_long *addr, u_short *port,
                   void *data, u_short size, u_long timeout)
{
    IPHDR *ip;
    UDPHDR *uh;
    NETBUF *nb;

    if(sock->so_rx_nb == 0) {
        slEventWait(&sock->so_rx_rdy, timeout);
    }

    if((nb = sock->so_rx_nb) == 0)
        return 0;
    sock->so_rx_nb = 0;

    ip = nb->nb_nw.vp;
    *addr = ip->ip_src;

    uh = nb->nb_tp.vp;
    *port = uh->uh_sport;

    if(size > nb->nb_ap.sz)
        size = nb->nb_ap.sz;

    memcpy(data, nb->nb_ap.vp, size);
    slNetBufFree(nb);

    return size;
}

/*!
 * \brief Close UDP socket.
 *
 * The memory occupied by the socket is immediately released
 * after calling this function. The application  must not use 
 * the socket after this call.
 *
 * \param sock Socket descriptor. This pointer must have been 
 *             retrieved by calling slUdpCreateSocket().
 *
 * \return 0 on success, -1 otherwise.
 */
int slUdpDestroySocket(UDPSOCKET *sock)
{
    UDPSOCKET *sp;
    UDPSOCKET **spp;
    int rc = -1;

    spp = &udpSocketList;
    sp = udpSocketList;

    while(sp) {
        if(sp == sock) {
            *spp = sock->so_next;
            slHeapFree(sock);
            rc = 0;
            break;
        }
        spp = &sp->so_next;
        sp = sp->so_next;
    }
    return rc;
}

/*!
 * \brief Find a matching socket.
 *
 * Loop through all sockets and find a matching one.
 *
 * \note Applications typically do not need to call this function.
 *
 * \param port  Local port number.
 *
 * \return Socket descriptor.
 */
UDPSOCKET *slUdpFindSocket(u_short port)
{
    UDPSOCKET *sp;
    UDPSOCKET *sock = 0;

    for(sp = udpSocketList; sp; sp = sp->so_next) {
        if(sp->so_local_port == port) {
            sock = sp;
            break;
        }
    }
    return sock;
}

/*@}*/
