
#include <sys/atom.h>
#include <sys/timer.h>
#include <sys/thread.h>
#include <sys/event.h>


/*!
 * \brief Timer callback in case of event timeout.
 *
 * \param timer Handle of the elapsed timeout timer.
 * \param arg   Handle of an event queue.
 *
 * \note This routine is running in interrupt context.
 */
void slEventTimeout(HANDLE timer, void *arg)
{
    NUTTHREADINFO *tqp;
    NUTTHREADINFO * volatile *tqpp = arg;

    /*
     * A signaled queue is an empty queue. So our
     * thread already left this queue.
     */
    if((tqp = *tqpp) != SIGNALED) {

        /*
         * Walk down the linked list and identify
         * the thread by the timer id it is waiting
         * for.
         */
        while(tqp) {
            if(tqp->td_timer == timer) {
                /*
                 * Found the thread. Remove it from the
                 * event queue and add it to the queue
                 * of threads, which are ready to run.
                 */
                *tqpp = tqp->td_qnxt;
                tqp->td_qnxt = 0;
                tqp->td_timer = 0;
                slThreadResumeAsync(tqp);
                break;
            }
            tqpp = &tqp->td_qnxt;
            tqp = tqp->td_qnxt;
        }
    }
}

/*!
 * \brief Wait for an event in a specified queue.
 *
 * Give up the CPU until another thread posts an event to this
 * queue or until a time-out occurs, whichever comes first.
 * If previously an event had been posted to this queue without 
 * any thread waiting, then the thread will continue running,
 * if no other thread with higher priority is ready to run.
 *
 * \param qhp Identifies the queue to wait on.
 * \param ms  Maximum wait time in milliseconds.
 *
 * \return 0 if event received, -1 on timeout.
 *
 * \note Timeout is limited to the granularity of the system timer.
 */
int slEventWait(volatile HANDLE *qhp, u_long ms)
{
    slEnterCritical();

    /*
     * Check for posts on a previously empty queue. 
     */
    if(*qhp == SIGNALED) {
        //if(runningThread == app1 || runningThread == app2 || runningThread == app3 || runningThread == app4)
        //    slPrintFormat(0, "(%04X-%04X)", runningThread, qhp);

        /*
         * Even if already signaled, switch to any other thread, which 
         * is ready to run and has the same or higher priority.
         */
        slExitCritical();
        slThreadYield();
        slEnterCritical();
        if(*qhp == SIGNALED)
            *qhp = 0;
        slExitCritical();
        return 0;
    }

    /*
     * Remove the current thread from the list
     * of running threads and add it to the
     * specified queue.
     */
#ifdef NUTDEBUG
    if(os_trace)
        slPrintFormat(0, "Rem<%04x>", runningThread);
#endif
    slThreadRemoveQueue(runningThread, &runQueue);
#ifdef NUTDEBUG
    if(os_trace)
        slKDumpThreadList();
        //slKDumpThreadQueue(runQueue);
#endif
    runningThread->td_state = TDS_SLEEP;
    //if(runningThread == app1 || runningThread == app2 || runningThread == app3 || runningThread == app4)
    //    slPrintFormat(0, "[%04X-%04X]", runningThread, qhp);
    slThreadAddPriQueue(runningThread, (NUTTHREADINFO **)qhp);

    /*
     * If a timeout value had been given, start
     * a oneshot timer.
     */
    if(ms)
        runningThread->td_timer = slTimerStart(ms, slEventTimeout, (void *)qhp, TM_ONESHOT);
    else
        runningThread->td_timer = 0;

    /*
     * Switch to the next thread, which is ready
     * to run.
     */
#ifdef NUTDEBUG
    if(os_trace)
        slPrintFormat(0, "SWW<%04x %04x>", runningThread, runQueue);
#endif
    slThreadSwitch();

    /*
     * If our timer handle is still set, we were
     * woken up by an event. If the handle is
     * cleared and a timeout value had been
     * specified, then we know that a timeout
     * occured.
     */
    if(runningThread->td_timer)
        runningThread->td_timer = 0;
    else if(ms) {
        slExitCritical();
        return -1;
    }
    slExitCritical();
    return 0;
}

/*!
 * \brief Wait for a new event in a specified queue.
 *
 * Give up the CPU until another thread posts an event to this
 * queue or until a time-out occurs, whichever comes first.
 * Previously posted events to an empty queue will be ignored.
 *
 * \param qhp Identifies the queue to wait on.
 * \param ms  Maximum wait time in milliseconds.
 *
 * \return 0 if event received, -1 on timeout.
 *
 * \note Timeout is limited to the granularity of the system timer.
 */
int slEventWaitNext(volatile HANDLE *qhp, u_long ms)
{
    int rc;

    slEnterCritical();

    /*
     * Check for posts on a previously empty queue. 
     */
    if(*qhp == SIGNALED)
        *qhp = 0;

    rc = slEventWait(qhp, ms);

    slExitCritical();

    return rc;
}

/*!
 * \brief Asynchronously post an event to a specified queue.
 *
 * Wake up the thread with the highest priority waiting on
 * the specified queue. But even if the priority of the woken
 * thread is higher than the current thread's priority, the
 * current one continues running.
 *
 * \note It is save to call this function from within an interrupt
 *       handler. In any case interrupts should be disabled when
 *       calling this function.
 *
 * \param qhp Identifies the queue an event is posted to.
 *
 */
void slEventPostAsync(HANDLE volatile *qhp)
{
    NUTTHREADINFO *td;

    if(*qhp != SIGNALED) {
        if((td = *qhp) != 0) {
            slThreadRemoveQueue(td, (NUTTHREADINFO *volatile *)qhp);

            /*
             * Stop any running timeout timer.
             */
            if(td->td_timer)
                slTimerStopAsync(td->td_timer);
            slThreadResumeAsync(td);
        }
        else 
            *qhp = SIGNALED;
    }
}

/*!
 * \brief Post an event to a specified queue.
 *
 * Wake up the thread with the highest priority waiting on
 * this queue. If the priority of the waiting thread is higher
 * or equal than the current thread's priority, then the
 * current thread is stopped and CPU control is passed
 * to the waiting thread.
 *
 * \param qhp Identifies the queue an event is posted to.
 *
 */
void slEventPost(HANDLE *qhp)
{
    slEnterCritical();

    slEventPostAsync(qhp);

    /*
     * If any thread with higher priority is
     * ready to run, switch the context.
     */
    if(runningThread != runQueue) {
        runningThread->td_state = TDS_READY;
#ifdef NUTDEBUG
        if(os_trace)
            slPrintFormat(0, "SWP<%04x %04x>", runningThread, runQueue);
#endif
        slThreadSwitch();
    }
    slExitCritical();
}

/*!
 * \brief Asynchronously broadcast an event to a specified queue.
 *
 * Wake up all threads waiting on this queue. But even if the
 * priority of any woken thread is higher than the current thread's
 * priority, the current one continues running.
 *
 * \note It is save to call this function from within an interrupt
 *       handler. In any case interrupts must be disabled when
 *       calling this function.
 *
 * \param qhp Identifies the queue an event is broadcasted to.
 *
 */
void slEventBroadcastAsync(HANDLE *qhp)
{
    if(*qhp != SIGNALED)
        while(*qhp)
            slEventPostAsync(qhp);
}

/*!
 * \brief Broadcast an event to a specified queue.
 *
 * Wake up all threads waiting on this queue. If the
 * priority of any waiting thread is higher or equal
 * than the current thread's priority, then the
 * current thread is stopped and CPU control is passed
 * to the woken up thread with the highest priority.
 *
 * \param qhp Identifies the queue an event is broadcasted to.
 *
 */
void slEventBroadcast(HANDLE *qhp)
{
    if(*qhp != SIGNALED)
        while(*qhp)
            slEventPost(qhp);
}

/*@}*/
