
#include <string.h>

#include <sys/atom.h>
#include <sys/heap.h>


/*!
 * \brief List of free nodes.
 */
HEAPNODE* volatile heapFreeList = 0;

/*!
 * \brief Number of bytes available.
 */
u_short available = 0;

/*!
 * \brief
 * Allocate a block from heap memory.
 *
 * This functions allocates a memory block of the specified
 * size and returns a pointer to that block.
 *
 * The actual size of the allocated block is larger than the
 * requested size because of space required for maintenance
 * information. This additional information is invisible to
 * the application.
 *
 * The routine looks for the smallest block that will meet
 * the required size and releases it to the caller. If the
 * block being requested is usefully smaller than the smallest
 * free block then the block from which the request is being
 * met is split in two. The unused portion is put back into
 * the free-list.
 *
 * The contents of the allocated block is unspecified.
 * To allocate a block with all bytes set to zero use
 * slHeapAllocClear().
 *
 * \note Do not use this function in interrupt routines.
 *
 * \param size Size of the requested memory block.
 *
 * \return Pointer to the allocated memory block if the
 *         function is successful or NULL if the requested
 *         amount of memory is not available.
 */
void *slHeapAlloc(u_short size)
{
    HEAPNODE *node;
    HEAPNODE **npp;
    HEAPNODE *fit = 0;
    HEAPNODE **fpp = 0;

    //slEnterCritical();
    if(size >= available) {
#ifdef NUTDEBUG
        if(heap_trace)
            slPrintString_P(0, PSTR("MEMOVR\r\n"));
#endif
        return 0;
    }

    /*
     * We need additional space in front of the allocated memory
     * block to store its size. If this is still less than the
     * space required by a free node, increase it.
     */
    if((size += 6) < sizeof(HEAPNODE))
        size = sizeof(HEAPNODE);

#ifdef NUTDEBUG
    if(heap_trace)
        slPrintFormat(0, "A(%d,", size);
#endif

    /*
     * Walk through the linked list of free nodes and find the best fit.
     */
    node = heapFreeList;
    npp = (HEAPNODE **)&heapFreeList;
    while(node) {

        /*
         * Found a note that fits?
         */
        if(node->hn_size >= size) {
            /*
             * If it's an exact match, we don't
             * search any further.
             */
            if(node->hn_size == size) {
                fit = node;
                fpp = npp;
                break;
            }

            /*
             * Is it the first one we found
             * or was the previous one larger?
             */
            if(fit == 0 || (fit->hn_size > node->hn_size)) {
                fit = node;
                fpp = npp;
            }
        }
        npp = &node->hn_next;
        node = node->hn_next;
    }

    if(fit) {
        /*
         * If the node we found is larger than the
         * required space plus the space needed for
         * a new node plus a defined threshold, then
         * we split it.
         */
        if(fit->hn_size > size + sizeof(HEAPNODE) + ALLOC_THRESHOLD) {
            node = (HEAPNODE *)((u_short)fit + size);
            node->hn_size = fit->hn_size - size;
            node->hn_next = fit->hn_next;
            fit->hn_size = size;
            *fpp = node;
        }
        else
            *fpp = fit->hn_next;

        available -= fit->hn_size;
        *((u_long *)(((char *)fit) + (fit->hn_size - 4))) = 0xDEADBEEF;
        fit = (HEAPNODE *)&fit->hn_next;
    }

#ifdef NUTDEBUG
    if(heap_trace) 
        slPrintFormat(0, "%x) ", (u_short)fit);
#endif
    //slExitCritical();
    return fit;
}

/*!
 * \brief Allocate an initialized block from heap memory.
 *
 * This functions allocates a memory block of the specified
 * size with all bytes initialized to zero and returns a
 * pointer to that block.
 *
 * \param size Size of the requested memory block.
 *
 * \return Pointer to the allocated memory block if the
 *         function is successful or NULL if the requested
 *         amount of memory is not available.
 */
void *slHeapAllocClear(u_short size)
{
    void *ptr;

    if((ptr = slHeapAlloc(size)) != 0)
        memset(ptr, 0, size);

    return ptr;
}

/*!
 * \brief Return a block to heap memory.
 *
 * An application calls this function, when a previously
 * allocated memory block is no longer needed.
 *
 * The heap manager checks, if the released block adjoins any
 * other free regions. If it does, then the adjacent free regions
 * are joined together to form one larger region.
 *
 * \note Do not use this function in interrupt routines.
 *
 * \param block Points to a memory block previously allocated
 *              through a call to slHeapAlloc().
 *
 * \return 0 on success, -1 if the caller tried to free
 *         a block which had been previously released.
 */
int slHeapFree(void *block)
{
    HEAPNODE *node;
    HEAPNODE **npp;
    HEAPNODE *fnode;

    //slEnterCritical();
#ifdef NUTDEBUG
    if(block) {
        u_short size;
        size = *(((u_short *)block) - 1);
        if(*((u_long *)(((char *)block) + (size - 6))) != 0xDEADBEEF)
            slPrintString(0, "\r\nMEMCORR-");

    }
    else
        slPrintString(0, "\r\nMEMNULL");
#endif

    /*
     * Convert our block into a node.
     */
    fnode = (HEAPNODE *)(((u_short *)block) - 1);

#ifdef NUTDEBUG
    if(heap_trace)
        slPrintFormat(0, "F(%d,%x) ", fnode->hn_size, (u_short)block);
#endif
    available += fnode->hn_size;

    /*
     * Walk through the linked list of free nodes and try
     * to link us in.
     */
    node = heapFreeList;
    npp = (HEAPNODE **)&heapFreeList;
    while(node) {
        /*
         * If there' s a free node in front of us, merge it.
         */
        if(((u_short)node + node->hn_size) == (u_short)fnode) {
            node->hn_size += fnode->hn_size;

            /*
             * If a free node is following us, merge it.
             */
            if(((u_short)node + node->hn_size) == (u_short)node->hn_next) {
                node->hn_size += node->hn_next->hn_size;
                node->hn_next = node->hn_next->hn_next;
            }
            break;
        }

        /*
         * If we walked past our address, link us to the list.
         */
        if((u_short)node > (u_short)fnode) {
            *npp = fnode;

            /*
             * If a free node is following us, merge it.
             */
            if(((u_short)fnode + fnode->hn_size) == (u_short)node) {
                fnode->hn_size += node->hn_size;
                fnode->hn_next = node->hn_next;
            }
            else
                fnode->hn_next = node;
            break;
        }

        /*
         * If we are within a free node, somebody tried
         * to free a block twice.
         */
        if(((u_short)node + node->hn_size) > (u_short)fnode) {
#ifdef NUTDEBUG
            if(heap_trace)
                slPrintString(0, "\r\nTWICE\r\n");
#endif
            //slExitCritical();
            return -1;
        }

        npp = &node->hn_next;
        node = node->hn_next;
    }

    /*
     * If no link was found, put us at the end of the list
     */
    if(!node) {
        fnode->hn_next = node;
        *npp = fnode;
    }

    //slExitCritical();
    return 0;
}

/*!
 * \brief
 * Add a new memory region to the free heap.
 *
 * This function is automatically called by sl/OS during
 * initialization.
 *
 * Applications typically do not call this function.
 *
 * \param addr Start address of the memory region.
 * \param size Number of bytes of the memory region.
 */
void slHeapAdd(void *addr, u_short size)
{
    *((u_short *)addr) = size;
    slHeapFree(((u_short *)addr) + 1);
}

/*!
 * \brief Return the number of bytes available.
 *
 * \return Number of bytes.
 */
u_short slHeapAvailable(void)
{
    return available;
}

/*@}*/
