
#include <sys/ifstream.h>

static char hexdigit[] = "0123456789ABCDEF";

/*!
 * \addtogroup xgIfStream
 */
/*@{*/

/*!
 * \brief Read from a stream device.
 *
 * \param dev  Identifies the device to read from. This pointer
 *             must have been retrieved by calling slDeviceOpen().
 * \param data Pointer to the buffer that receives the data. Set to
 *             null to flush the input buffer.
 * \param size Size of the buffer. Ignored if the data pointer is
 *             null.
 *
 * \return The number of bytes read, 0 on timeouts and broken
 *         connections or -1 in case of an error.
 */
int slIfStreamRead(NUTDEVICE *dev, char *data, int size)
{
    int rc;
    IFSTREAM *ifs;

    /*
     * Check for valid device pointer.
     */
    if((void *)dev < RAMSTART)
        return -1;

    /*
     * Check if input is supported.
     */
    ifs = (IFSTREAM *)dev->dev_icb;
    if(ifs->if_input == 0)
        return -1;

    /*
     * Call without data pointer discards receive buffer.
     */
    if(data == 0) {
        ifs->if_rd_idx = ifs->if_rx_idx;
        return 0;
    }

    /*
     * Check for valid buffer pointer.
     */
    if((void *)data < RAMSTART)
        return -1;

    /*
     * If buffer empty, wait for input. 
     * Return 0 in case of timeout.
     */
    if(ifs->if_rd_idx == ifs->if_rx_idx)
        if((*ifs->if_input)(dev))
            return 0;

    /*
     * Get characters from receive buffer.
     */
    for(rc = 0; rc < size; rc++) {
        if(ifs->if_rd_idx == ifs->if_rx_idx)
            break;
        *data++ = ifs->if_rx_buf[ifs->if_rd_idx++];
    }
    return rc;
}

/*!
 * \brief Read from a stream device translating non printables.
 *
 * This call works like slIfStreamRead(), but translates
 * character codes below 33, above 126 or equal 34 into
 * their hex representation prepended by a backslash and the
 * letter x. Backslashes are duplicated and the complete string
 * is terminated by character code zero. This offers the ability 
 * to store binary data in printable strings.
 *
 * \todo Very specific code, should be moved to an extra module
 *       to save ROM space of ordinary applications.
 *
 * \param dev  Identifies the device to read from. This pointer
 *             must have been retrieved by calling slDeviceOpen().
 * \param data Pointer to the buffer that receives the translated string.
 * \param size Size of the buffer.
 *
 * \return The number of bytes read or -1 in case of an error. Due
 *         to code translation, this may be less than the number of
 *         characters returned in the string.
 */
int slIfStreamReadTran(NUTDEVICE *dev, char *data, int size)
{
    int rc = 0;
    int sc;
    u_char ch;
    IFSTREAM *ifs;
    
    /*
     * Check if input is supported.
     */
    ifs = (IFSTREAM *)dev->dev_icb;
    if(ifs->if_input == 0)
        return -1;

    /*
     * Call without data pointer discards receive buffer.
     */
    if(data == 0 || size == 0) {
        ifs->if_rd_idx = ifs->if_rx_idx;
        return 0;
    }

    /*
     * If buffer empty, wait for input.
     * Return 0 on timeout.
     */
    if(ifs->if_rd_idx == ifs->if_rx_idx)
        if((*ifs->if_input)(dev))
            return 0;

    /*
     * Get characters from receive buffer.
     */
    size--;
    for(sc = 0; sc < size; ) {
        if(ifs->if_rd_idx == ifs->if_rx_idx)
            break;
        ch = ifs->if_rx_buf[ifs->if_rd_idx];
        if(ch < 33 || ch > 126  || ch == 34) {
            if(sc + 4 >= size)
                break;
            *data++ = '\\';
            *data++ = 'x';
            *data++ = hexdigit[ch >> 4];
            *data++ = hexdigit[ch & 0x0f];
            sc += 4;
        }
        else {
            if(ch == '\\') {
                if(sc + 2 >= size)
                    break;
                *data++ = '\\';
                sc++;
            }
            *data++ = ch;
            sc++;
        }
        ifs->if_rd_idx++;
        rc++;
    }
    *data = 0;

    return rc;
}

/*!
 * \brief Get a line from a stream device.
 *
 * Reads a string from the specified device. Characters are
 * read up to and including the first end of line character
 * or until the number of characters read is equal to the
 * specified maximum or until a timeout occurs, whichever 
 * comes first.
 *
 * \param dev  Identifies the device to read from. This pointer
 *             must have been retrieved by calling slDeviceOpen().
 * \param data Points to the buffer that receives the zero 
 *             terminated string. End of line characters are not 
 *             stored.
 * \param size Specifies the size of the buffer. The maximum
 *             number of characters read is one less because
 *             of the terminating null character.
 *
 * \return The number of bytes read or -1 in case of an error.
 *         Receiving no character is being considered an error.
 */
int slIfStreamGetLine(NUTDEVICE *dev, char *data, int size)
{
    int rc;
    IFSTREAM *ifs;
    u_char ch;

    /*
     * Check for valid device pointer.
     */
    if((void *)dev < RAMSTART)
        return -1;

    /*
     * Check if input is supported.
     */
    ifs = (IFSTREAM *)dev->dev_icb;
    if(ifs->if_input == 0)
        return -1;

    /*
     * Check for valid buffer pointer.
     */
    if((void *)data < RAMSTART)
        return -1;

    for(rc = 0; rc < size - 1; ) {
        /*
         * If buffer is empty, call input handler.
         */
        if(ifs->if_rd_idx == ifs->if_rx_idx)
            if((*ifs->if_input)(dev)) {
                if(rc == 0)
                    rc = -1;
                break;
            }

        ch = ifs->if_rx_buf[ifs->if_rd_idx++];
        if(ch == '\r' || ch == '\n') {
            if(ifs->if_last_eol == 0 || ifs->if_last_eol == ch) {
                ifs->if_last_eol = ch;
                break;
            }
        }
        else {
            ifs->if_last_eol = 0;
            *data++ = ch;
            rc++;
        }
    }
    *data = 0;

    return rc;
}


/*!
 * \brief Write to a stream device.
 *
 * The data isn't immediately transfered to the physical device, 
 * but buffered in a transmit buffer. Transmission starts either 
 * when the buffer is full or when this function is called with
 * a zero data pointer.
 *
 * \param dev  Identifies the device to write to. This pointer
 *             must have been retrieved by calling slDeviceOpen().
 * \param data Pointer to data to be written. If this pointer is
 *             NULL, the stream device starts transmitting the
 *             buffered data.
 * \param len  Number of bytes to write.
 *
 * \return The number of bytes written or -1 in case of an error.
 */
int slIfStreamWrite(NUTDEVICE *dev, CONST char *data, int len)
{
    int rc;
    IFSTREAM *ifs;
    u_char ix;

    /*
     * Check for valid device pointer.
     */
    if((void *)dev < RAMSTART)
        return -1;

    /*
     * Check if output is supported.
     */
    ifs = dev->dev_icb;
    if(ifs->if_output == 0)
        return -1;

    /*
     * Call without data pointer starts transmission.
     */
    if(data == 0) {
        if((*ifs->if_output)(dev) < 0)
            return -1;
        return 0;
    }

    /*
     * Check for valid buffer pointer.
     */
    if((void *)data < RAMSTART)
        return -1;

    /*
     * Start transmission if output buffer full.
     */
    if((ix = ifs->if_wr_idx + 1) == ifs->if_tx_idx)
        if((*ifs->if_output)(dev) < 0)
            return -1;

    /*
     * Put characters in transmit buffer.
     */
    for(rc = 0; rc < len; rc++) {
        if(ix == ifs->if_tx_idx)
            break;
        ifs->if_tx_buf[ifs->if_wr_idx] = *data++;
        ifs->if_wr_idx = ix++;
    }
    return rc;
}

/*!
 * \brief Translate and write a string to a stream device.
 *
 * This call works like slIfStreamWrite(), but translates
 * any backslash followed by the letter x followed by two
 * hexadecimal uppercase digits into the binary representation. 
 * Simple backslashes are expected to be duplicated and the complete 
 * string must be terminated by character code zero. This offers the 
 * ability to write binary data from a printable strings.
 *
 * The data isn't immediately transfered to the physical device, 
 * but buffered in a transmit buffer. Transmission starts either 
 * when the buffer is full or when this function is called with
 * a zero data pointer.
 *
 * \param dev  Identifies the device to write to. This pointer
 *             must have been retrieved by calling slDeviceOpen().
 * \param data Pointer to a zero terminated string to be translated
 *             and written. If this pointer is NULL, the stream device 
 *             starts transmitting the buffered data.
 *
 * \return The number of characters written or -1 in  case of an error. 
 *         Due to code translation, this may be less than the number
 *         of characters contained in the string.
 */
int slIfStreamWriteTran(NUTDEVICE *dev, CONST char *data)
{
    int rc = 0;
    IFSTREAM *ifs;
    u_char ix;
    u_char ch;

    /*
     * Check if output is supported.
     */
    ifs = dev->dev_icb;
    if(ifs->if_output == 0)
        return -1;

    /*
     * Call without data pointer starts transmission.
     */
    if(data == 0) {
        if((*ifs->if_output)(dev) < 0)
            return -1;
        return 0;
    }

    /*
     * Start transmission if output buffer full.
     */
    if(ifs->if_wr_idx + 1 == ifs->if_tx_idx) 
        if((*ifs->if_output)(dev) < 0 )
            return -1;

    /*
     * If the buffer is still full, we have write timeout.
     */
    if((ix = ifs->if_wr_idx + 1) == ifs->if_tx_idx) 
        return 0;

    /*
     * Put characters in transmit buffer.
     */
    while(*data) {
        if(ix == ifs->if_tx_idx)
            break;
        if(*data == '\\') {
            if(*++data == 'x') {
                data++;
                ch = ((*data > '9') ? *data - 55 : *data - 48) << 4;
                data++;
                ch += (*data > '9') ? *data - 55 : *data - 48;
                data++;
            }
            else {
                ch = *data++;
            }
        }
        else {
            ch = *data++;
        }
        ifs->if_tx_buf[ifs->if_wr_idx] = ch;
        ifs->if_wr_idx = ix++;
        rc++;
    }
    return rc;
}

/*!
 * \brief Write program space data to a stream device.
 *
 * \param dev  Identifies the device to write to. This pointer
 *             must have been retrieved by calling slDeviceOpen().
 * \param data Pointer to data in program space to be written.
 * \param len  Number of bytes to write.
 *
 * \return The number of bytes written or -1 in case of an error.
 */
int slIfStreamWrite_P(NUTDEVICE *dev, PGM_P data, int len)
{
    int rc;
    IFSTREAM *ifs;
    u_char ix;

    /*
     * Check for valid device pointer.
     */
    if((void *)dev < RAMSTART)
        return -1;

    /*
     * Check if output is supported.
     */
    ifs = (IFSTREAM *)dev->dev_icb;
    if(ifs->if_output == 0)
        return -1;

    /*
     * If buffer full, wait for output.
     */
    if((ix = ifs->if_wr_idx + 1) == ifs->if_tx_idx)
        if((*ifs->if_output)(dev) < 0)
            return -1;

    /*
     * Put characters in transmit buffer.
     */
    for(rc = 0; rc < len; rc++) {
        if(ix == ifs->if_tx_idx)
            break;
        ifs->if_tx_buf[ifs->if_wr_idx] = PRG_RDB(data);
        data++;
        ifs->if_wr_idx = ix++;
    }
    return rc;
}

/*!
 * \brief Flush stream device buffer.
 *
 * Flushes the output buffer for the specified device by
 * writing its current contents to the corresponding device.
 *
 * \param dev  Identifies the device to flush. This pointer
 *             must have been retrieved by directly or indirectly
 *             calling slDeviceOpen().
 *
 * \return 0 on success, -1 on failures.
 */
int slIfStreamFlush(NUTDEVICE *dev)
{
    IFSTREAM *ifs;

    /*
     * Check for valid device pointer.
     */
    if((void *)dev >= RAMSTART) {
        if((void *)(ifs = (IFSTREAM *)dev->dev_icb) >= RAMSTART && ifs->if_flush)
            return (*ifs->if_flush)(dev);
    }
    return -1;
}

/*@}*/
