
#include <stdarg.h>
#include <string.h>

#include <sys/atom.h>
#include <sys/print.h>
#include <sys/kprint.h>

/*!
 * \brief Write a buffer to the debug output.
 *
 * The contents of the buffer is immediately sent to
 * the output device. The function will not return
 * unless all characters are sent.
 *
 * \param data Buffer to be written.
 * \param len  Number of characters to be printed.
 *
 * \return The number of characters printed or -1 in case
 *         of an error.
 */
int slKPrintBinary(CONST char *data, int len)
{
    int rc = 0;

    slEnterCritical();
    while(len) {
        loop_until_bit_is_set(USR, 5);
        outp(*data, UDR);
        data++;
        rc++;
        len--;
    }
    slExitCritical();
    return rc;
}

int slKPrintBinary_P(PGM_P data, int len)
{
    int rc = 0;

    slEnterCritical();
    while(len) {
        loop_until_bit_is_set(USR, 5);
        outp(PRG_RDB(data), UDR);
        data++;
        rc++;
        len--;
    }
    slExitCritical();
    return rc;
}

/*!
 * \brief Print a string on the debug output.
 *
 * The contents of the string is immediately sent to
 * the output device. The function will not return
 * unless all characters are sent.
 *
 * \param str String to be printed.
 *
 * \return The number of characters printed or -1 in case
 *         of an error.
 */
int slKPrintString(CONST char *str)
{
    return slKPrintBinary(str, strlen(str));
}

int slKPrintString_P(PGM_P str)
{
    return slKPrintBinary_P(str, strlen_P(str));
}

/*!
 * \brief Print a numeric value on the debug output.
 *
 * The numeric value will be converted to an ASCII string and 
 * printed to the debug output device. 
 *
 * The resulting string is immediately sent to
 * the output device. The function will not return
 * unless all characters are sent.
 *
 * \param val   Value to be printed.
 * \param radix Number base, may be any value between 2 and 16.
 * \param width Minimum width or 0 if no justification required.
 * \param flags Format flags. Any of the following may be ored:
 * - FMTFLG_ZERO for output padding with zeros.
 * - FMTFLG_PLUS forces printing a sign character.
 * - FMTFLG_MINUS for left justification.
 * - FMTFLG_SIGNED for signed values.
 * - FMTFLG_CAPITAL for using capital letters with base above 10.
 *
 * \return The number of characters printed or -1 in case of an error.
 *
 * @see Format flags
 *
 */
int slKPrintInteger(u_long val, u_char radix, u_char width, u_char flags)
{
    u_char *digits;
    u_char raw[16];
    int result;
    u_char rem;
    u_char sign;
    u_char cnt;

    if(radix < 2)
        radix = 10;
    result = 0;

    sign = 0;
    if(flags & FMTFLG_PLUS)
        sign = '+';
    if(flags & FMTFLG_SIGNED)
        if((long)val < 0) {
            sign = '-';
            val = -(long)val;
        }

    /*
     * Fill scratch buffer with raw digits.
     */
    digits = (flags & FMTFLG_CAPITAL) ? "0123456789ABCDEF"
                                      : "0123456789abcdef";
    cnt = 0;
    do {
        rem = (u_char)(val % (u_long)radix);
        val = val / (u_long)radix;
        raw[cnt++] = digits[rem];
    } while(val);

    /*
     * Calculate the remaining width for padding
     */
    if(width > cnt) {
        if(sign)
            width--;
        width -= cnt;
    }
    else
        width = 0;

    /*
     * If not zero padded and not left justified,
     * we put enough spaces in front.
     */
    if((flags & (FMTFLG_ZERO | FMTFLG_MINUS)) == 0)
        while(width) {
            slKPrintBinary(" ", 1);
            result++;
            width--;
        }
    if(sign) {
        slKPrintBinary(&sign, 1);
        result++;
    }
    if((flags & (FMTFLG_ZERO | FMTFLG_MINUS)) == FMTFLG_ZERO) {
        result += width;
        while(width) {
            slKPrintBinary("0", 1);
            width--;
        }
    }
    while(cnt-- > 0) {
        slKPrintBinary(&raw[cnt], 1);
        result++;
    }
    while(width) {
        slKPrintBinary(" ", 1);
        result++;
        width--;
    }
    return result;
}

/*!
 * \brief Print parameters using a format string.
 *
 * This function formats and prints characters and values
 * to a specified character output routine.
 *
 * The format string consists of ordinary characters, escape
 * sequences, and format specifications. Ordinary characters
 * and escape sequences are copied to the output routine in
 * the order of their appearance. The format is scanned from
 * left to right. If a format specification is encountered,
 * the value of the corresponding argument is converted and
 * output according to that format specification. If there
 * are more arguments than format specifications, the extra
 * arguments are ignored. If there are more format specifications
 * than arguments, the results are undefined.
 *
 * Recognized format specifiers are:
 * \arg %%c prints a character.
 * \arg %%s prints a string.
 * \arg %%o prints a numerical value in octal digits.
 * \arg %%d prints a signed numerical value in deciaml digits.
 * \arg %%u prints an unsigned numerical value in decimal digits.
 * \arg %%x prints a numerical value in hexadecimal digits using lowercase letters.
 * \arg %%X prints a numerical value in hexadecimal digits using uppercase letters.
 *
 * The resulting string is immediately sent to
 * the output device. The function will not return
 * unless all characters are sent. For buffered
 * output use slPrintFormat().
 *
 * \param fmt Format string.
 *
 * \return The number of characters printed, or a negative value
 *         in case of an output error.
 */
int slKPrintFormat(CONST char *fmt, ...)
{
    va_list ap;
    u_char *s;
    u_char *cp;
    int width;
    int result;
    u_short len;
    u_long val;
    u_char radix;
    u_char flags;
    u_char isize;

    va_start(ap, fmt);
    for(result = 0; *fmt; ++fmt) {
        if(*fmt != '%') {
            if(slKPrintBinary(fmt, 1) < 0)
                return -1;
            result++;
            continue;
        }

        flags = 0;
        while(1) {
            if(*++fmt == '-')
                flags |= FMTFLG_MINUS;
            else if(*fmt == '+')
                flags |= FMTFLG_PLUS;
            else if(*fmt == '0')
                flags |= FMTFLG_ZERO;
            else
                break;
        }

        if(*fmt == '*') {
            ++fmt;
            width = va_arg(ap, int);
            if(width < 0) {
                width = -width;
                flags |= FMTFLG_MINUS;
            }
        }
        else {
            width = 0;
            while(*fmt >= '0' && *fmt <= '9')
                width = width * 10 + (*fmt++ - '0');
        }

        isize = 's';
        if(*fmt == 'l') {
            isize = *fmt;
            ++fmt;
        }

        if(*fmt == 'c' || *fmt == 's') {
            s = 0;
            if(*fmt == 's') {
                if((s = va_arg(ap, char *)) == 0) {
                    s = "(NULL)";
                    len = 6;
                }
                else {
                    len = 0;
                    cp = s;
                    while(*cp++)
                        len++;
                }
            }
            else
                len = 1;
            if((flags & FMTFLG_MINUS) == 0)
                while(width > len) {
                    slKPrintBinary(" ", 1);
                    result++;
                    width--;
                }
            if(s) {
                slKPrintBinary(s, len);
                result += len;
            }
            else {
                u_char ch = (u_char)va_arg(ap, int);
                slKPrintBinary(&ch, 1);
                result++;
            }
            while(width > len) {
                slKPrintBinary(" ", 1);
                result++;
                width--;
            }
            continue;
        }

        radix = 10;
        if(*fmt == 'o')
            radix = 8;
        else if(*fmt == 'X' || *fmt == 'x') {
            if(*fmt == 'X')
                flags |= FMTFLG_CAPITAL;
            radix = 16;
        }
        else if(*fmt == 'd') 
            flags |= FMTFLG_SIGNED;
        else if(*fmt != 'u') {
            if(*fmt != '%') {
                slKPrintBinary("%", 1);
                result++;
            }
            if(*fmt) {
                slKPrintBinary(fmt, 1);
                result++;
            }
            else
                fmt--;
            continue;
        }
        if(isize == 'l') {
            if(flags & FMTFLG_SIGNED)
                val = va_arg(ap, long);
            else
                val = va_arg(ap, unsigned long);
        }
        else {
            if(flags & FMTFLG_SIGNED)
                val = va_arg(ap, int);
            else
                val = va_arg(ap, unsigned int);
        }
        result += slKPrintInteger(val, radix, width, flags);
    }
    va_end(ap);

    return result;
}

/*@}*/
