
#include <string.h>

#include <sys/device.h>

#include <sys/timer.h>
#include <sys/heap.h>

#include <arpa/inet.h>
#include <net/if_var.h>
#include <netinet/sostream.h>

#include <sys/print.h>

#include <fs/uromfs.h>
#include "dencode.h"

#include <pro/httpd.h>

/*!
 * \addtogroup xgHTTPD
 */
/*@{*/

static struct {
    char* ext;
    char* type;
} mimeTypes[] = {
    { ".txt",  "text/plain" },
    { ".html", "text/html" },
    { ".gif",  "image/gif" },
    { ".jpg",  "image/jpeg" }
};


/*!
 * \brief Send top lines of a standard HTML header.
 *
 * Sends HTTP and Server version lines.
 *
 * \param sostream Socket stream of the connection, previously
 *                 created by calling slSoStreamCreate().
 * \param req      The associated client request.
 * \param status   Response status, error code or 200, if no
 *                 error occured.
 * \param title    Error text, or OK, if no error occured.
 */
void slHttpSendHeaderTop(NUTDEVICE *sostream, REQUEST *req, int status, char* title)
{
    slPrintFormat(sostream, "HTTP/%d.%d %d %s\r\n",
                req->req_version / 10, req->req_version % 10, status, title);
    slPrintString(sostream, "Server: Ethernut 2.5\r\n");
}

/*!
 * \brief Send bottom lines of a standard HTML header.
 *
 * Sends Content-Type, Content-Lenght and Connection lines.
 *
 * \param sostream  Socket stream of the connection, previously
 *                  created by calling slSoStreamCreate().
 * \param mime_type Points to a string that specifies the content
 *                  type. Examples are "text/html", "image/png", 
 *                  "image/gif", "video/mpeg" or "text/css".
 *                  A null pointer is ignored.
 * \param bytes     Content length of the data following this
 *                  header. Ignored, if negative.
 */
void slHttpSendHeaderBot(NUTDEVICE *sostream, char* mime_type, int bytes)
{
    if(mime_type)
        slPrintFormat(sostream, "Content-Type: %s\r\n", mime_type);
    if(bytes >= 0)
        slPrintFormat(sostream, "Content-Length: %d\r\n", bytes);
    slPrintFormat(sostream, "Connection: close\r\n\r\n");
}

static char *err_file =
    "<HTML>"
    "<HEAD>"
    "<TITLE>%d %s</TITLE>"
    "</HEAD>"
    "<BODY>"
    "%d %s"
    "</BODY>"
    "</HTML>";

/*!
 * \brief Send a HTTP error response.
 *
 * A canned error file is used.
 *
 * \param sostream Socket stream of the connection, previously
 *                 created by calling slSoStreamCreate().
 * \param req      Contains the HTTP request.
 * \param status   Error code to be returned.
 */
void slHttpSendError(NUTDEVICE *sostream, REQUEST *req, int status)
{
    char *title;

    switch(status) {
    case 400:
        title = "Bad Request";
        break;
    case 401:
        title = "Unauthorized";
        break;
    case 404:
        title = "Not Found";
        break;
    case 500:
        title = "Internal Error";
        break;
    case 501:
        title = "Not Implemented";
        break;
    default:
        title = "Error";
        break;
    }

    slHttpSendHeaderTop(sostream, req, status, title);
    if(status == 401) {
        char *cp = 0;
        char *realm = req->req_url;

        if((cp = strrchr(realm, '/')) != 0)
            *cp = 0;
        else
            realm = ".";
        slPrintFormat(sostream, "WWW-Authenticate: Basic realm=\"%s\"", realm);
        if(cp)
            *cp = '/';
    }
    slHttpSendHeaderBot(sostream, "text/html", -1);
    slPrintFormat(sostream, err_file, status, title, status, title);

}

/*!
 * \brief Return the mime type description of a specified file name.
 *
 * The mime type returned is based on the file extension.
 *
 * \todo Function to register additional mime types. Currently only
 *       .txt, .html, .gif and .jpg are supported.
 *
 * \param name Name of the file.
 *
 * \return A pointer to a static string, containing the
 *         associated mime type description. If the extension
 *         is not registered, "text/plain; charset=iso-8859-1"
 *         is returned. If the filename is empty, then
 *         "text/hatml; charset=iso-8859-1" is returned.
 */
char *slGetMimeType(char* name)
{
    int i;
    int fl;

    if(name == 0 || (fl = strlen(name)) == 0)
        return mimeTypes[1].type;    
    for(i = 0; i < sizeof(mimeTypes) / sizeof(*mimeTypes); i++)
        if(strcasecmp(&(name[fl - strlen(mimeTypes[i].ext)]), mimeTypes[i].ext) == 0)
            return mimeTypes[i].type;
    return mimeTypes[0].type;
}

/*!
 *
 */
static void slHttpProcessFileRequest(NUTDEVICE *sostream, REQUEST *req)
{
    ROMFILE *fd;
    int n;
    char *data;
    int size;
    int written;

    /*
     * Validate authorization.
     */
    if(slHttpAuthValidate(req)) {
        slHttpSendError(sostream, req, 401);
        return;
    }

    /*
     * Process CGI.
     */
    if(strncasecmp(req->req_url, "cgi-bin/", 8) == 0) {
        slCgiProcessRequest(sostream, req);
        return;
    }

    /*
     * Process file.
     */
    if((fd = slRomFileOpen(req->req_url)) == 0) {
        u_char *index;
        u_short urll;

        urll = strlen(req->req_url);
        if((index = slHeapAllocClear(urll + 12)) == 0) {
            slHttpSendError(sostream, req, 500);
            return;
        }
        if(urll)
            strcpy(index, req->req_url);
        if(urll && index[urll - 1] != '/')
            strcat(index, "/");
        strcat(index, "index.html");
        if((fd = slRomFileOpen(index)) == 0) {
            slHeapFree(index);
            slHttpSendError(sostream, req, 404);
            return;
        }
        slHeapFree(index);
    }
    size = slRomFileSize(fd);
    slHttpSendHeaderTop(sostream, req, 200, "Ok");
    slHttpSendHeaderBot(sostream, slGetMimeType(req->req_url), size);
    if(req->req_method != METHOD_HEAD) {
        if(size > 512)
            size = 512;
        if((data = slHeapAlloc(size)) != 0) {
            while((n = slRomFileRead(fd, data, size)) != 0) {
                written = slPrintBinary(sostream, data, n);
            }
            slHeapFree(data);
            slPrintFlush(sostream);
        }
    }
    slRomFileClose(fd);
}

/*!
 *
 */
static char *NextWord(char *str)
{
    while(*str && *str != ' ' && *str != '\t')
        str++;
    if(*str)
        *str++ = 0;
    while(*str == ' ' || *str == '\t')
        str++;
    return str;
}

/*!
 *
 */
static REQUEST *CreateRequestInfo(void)
{
    return slHeapAllocClear(sizeof(REQUEST));
}

/*!
 *
 */
static void DestroyRequestInfo(REQUEST *req)
{
    if(req->req_url)
        slHeapFree(req->req_url);
    if(req->req_query)
        slHeapFree(req->req_query);
    if(req->req_type)
        slHeapFree(req->req_type);
    if(req->req_cookie)
        slHeapFree(req->req_cookie);
    if(req->req_auth)
        slHeapFree(req->req_auth);
    if(req->req_agent)
        slHeapFree(req->req_agent);
    slHeapFree(req);
}

/*!
 * \brief Process the next HTTP request.
 *
 * Waits for the next HTTP request on an established connection
 * and processes it.
 *
 * \param sostream Socket stream of the connection, previously
 *                 created by calling slSoStreamCreate().
 */
void slHttpProcessRequest(NUTDEVICE *sostream)
{
    REQUEST *req;
    char *method;
    char *path;
    char *line;
    char *protocol;
    char *cp;

    /*
     * Parse the first line of the request.
     * <method> <path> <protocol>
     */
    if((req = CreateRequestInfo()) == 0)
        return;
    if((method = slHeapAlloc(256)) == 0) {
        DestroyRequestInfo(req);
        return;
    }
    slDeviceGetLine(sostream, method, 256);

    /*
     * Parse remaining request headers.
     */
    if((line = slHeapAlloc(256)) == 0) {
        slHeapFree(method);
        DestroyRequestInfo(req);
        return;
    }
    for(;;) {
        if(slDeviceGetLine(sostream, line, 256) <= 0)
            break;
        if(*line == 0)
            break;
        if(strncasecmp(line, "Authorization:", 14) == 0) {
            if(req->req_auth == 0) {
                cp = &line[14];
                while(*cp == ' ' || *cp == '\t')
                    cp++;
                if((req->req_auth = slHeapAlloc(strlen(cp) + 1)) == 0)
                    break;
                strcpy(req->req_auth, cp);
            }
        }
        else if(strncasecmp(line, "Content-Length:", 15) == 0)
            req->req_length = 0 /*atoi(&line[15])*/;
        else if(strncasecmp(line, "Content-Type:", 13) == 0) {
            if(req->req_type == 0) {
                cp = &line[13];
                while(*cp == ' ' || *cp == '\t')
                    cp++;
                if((req->req_type = slHeapAlloc(strlen(cp) + 1)) == 0)
                    break;
                strcpy(req->req_type, cp);
            }
        }
        else if(strncasecmp(line, "Cookie:", 7) == 0) {
            if(req->req_cookie == 0) {
                cp = &line[7];
                while(*cp == ' ' || *cp == '\t')
                    cp++;
                if((req->req_cookie = slHeapAlloc(strlen(cp) + 1)) == 0)
                    break;
                strcpy(req->req_cookie, cp);
            }
        }
        else if(strncasecmp(line, "User-Agent:", 11) == 0) {
            if(req->req_agent == 0) {
                cp = &line[11];
                while(*cp == ' ' || *cp == '\t')
                    cp++;
                if((req->req_agent = slHeapAlloc(strlen(cp) + 1)) == 0)
                    break;
                strcpy(req->req_agent, cp);
            }
        }
    }
    slHeapFree(line);

    path = NextWord(method);
    protocol = NextWord(path);
    NextWord(protocol);

    if(strcasecmp(method, "GET") == 0)
        req->req_method = METHOD_GET;
    else if(strcasecmp(method, "HEAD") == 0)
        req->req_method = METHOD_HEAD;
    else if(strcasecmp(method, "POST") == 0)
        req->req_method = METHOD_POST;
    else {
        slHeapFree(method);
        DestroyRequestInfo(req);
        slHttpSendError(sostream, req, 501);
        return;
    }
    if(*path == 0 || *protocol == 0) {
        slHeapFree(method);
        DestroyRequestInfo(req);
        slHttpSendError(sostream, req, 400);
        return;
    }

    req->req_version = 10;
    if(strcasecmp(protocol, "HTTP/0.9") == 0)
        req->req_version = 9;

    if((cp = strchr(path, '?')) != 0) {
        *cp++ = 0;
        if((req->req_query = slHeapAlloc(strlen(cp) + 1)) == 0) {
            slHeapFree(method);
            DestroyRequestInfo(req);
            return;
        }
        strcpy(req->req_query, cp);
    }
    if((req->req_url = slHeapAlloc(strlen(path) + 1)) == 0) {
        slHeapFree(method);
        DestroyRequestInfo(req);
        return;
    }
    strcpy(req->req_url, path);

    slHeapFree(method);

    if(slDecodePath(req->req_url) == 0)
        slHttpSendError(sostream, req, 400);
    else
        slHttpProcessFileRequest(sostream, req);

    DestroyRequestInfo(req);
}

/*@}*/
